from typing import Any, Dict
from pathlib import Path
from dataclasses import dataclass

import pytest

from wunderkafka.tests import TestConsumer, TestHTTPClient
from wunderkafka.serdes.avro import FastAvroDeserializer, ConfluentClouderaHeadersHandler
from wunderkafka.tests.consumer import Message
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient
from wunderkafka.consumers.constructor import HighLevelDeserializingConsumer


@dataclass
class Msg(object):
    payload: bytes
    deserialized: Dict[str, Any]

    def serialized(self, header: bytes) -> bytes:
        return header + self.payload


SIGNAL_MESSAGE = Msg(
    payload=b'\x08test\x0cstring\x0cstring\x04NA\xcc\xb8\xeb\xa6\x80_',
    deserialized={
        'source': 'test',
        'type': 'string',
        'id': 'string',
        'value': 'NA',
        'ts': 1632128298534,
    },
)


HEADERS = (
    b'\x01\x00\x00\x00\x00\x00\x00\x06\x9c\x00\x00\x00\x01',
    b'\x02\x00\x00\x00\x00\x00\x00\x08<',
    b'\x03\x00\x00\x08<',
)


@pytest.mark.parametrize("header", list(HEADERS))
def test_consume_moving_parts(sr_root: Path, topic: str, header: bytes) -> None:
    consumer = HighLevelDeserializingConsumer(
        consumer=TestConsumer([Message(topic, value=SIGNAL_MESSAGE.serialized(header))]),
        schema_registry=ClouderaSRClient(TestHTTPClient(sr_root), SimpleCache()),
        headers_handler=ConfluentClouderaHeadersHandler().parse,
        deserializer=FastAvroDeserializer(),
    )

    consumer.subscribe([topic], from_beginning=True)

    msgs = consumer.consume()
    [message] = msgs
    assert message.key() is None
    assert message.value() == SIGNAL_MESSAGE.deserialized



import pytest

from wunderkafka.structures import SRMeta
from wunderkafka.serdes.avro.headers import ConfluentClouderaHeadersHandler

parse = ConfluentClouderaHeadersHandler().parse
pack = ConfluentClouderaHeadersHandler().pack

BAD_HEADERS = (
    b'\x00',
    b'\x00\x00\x00=',
    b'\x08\xf2',
)


@pytest.mark.parametrize("bad_header", list(BAD_HEADERS))
def test_short_message(bad_header: bytes) -> None:
    with pytest.raises(RuntimeError):
        parse(bad_header)


# ToDo (tribunsky.kir): bring to common exceptions.
def test_unknown_protocol() -> None:
    with pytest.raises(RuntimeError):
        parse(b'\x05\x00\x00\x00=')
    with pytest.raises(ValueError):
        pack(5, meta=SRMeta(1, 1, 1))


def test_handle_hw_int() -> None:
    hdr = parse(b'\x03\x00\x00\x00=')
    assert hdr.protocol_id == 3
    assert hdr.schema_id == 61


def test_handle_hw_int2() -> None:
    serialized = b'\x03\x00\x00\x034'
    hdr = parse(serialized)
    assert hdr.protocol_id == 3
    assert hdr.schema_id == 820
    packed = pack(hdr.protocol_id, meta=SRMeta(hdr.schema_id, hdr.schema_version, hdr.meta_id))
    assert packed == serialized


def test_handle_hw_meta() -> None:
    hdr = parse(b'\x01\x00\x00\x00\x00\x00\x00\x00 \x00\x00\x00\x01')
    assert hdr.protocol_id == 1
    assert hdr.meta_id == 32
    assert hdr.schema_version == 1


def test_handle_hw_meta2() -> None:
    serialized = b'\x01\x00\x00\x00\x00\x00\x00\x01\xa8\x00\x00\x00\x01'
    hdr = parse(serialized)
    assert hdr.protocol_id == 1
    assert hdr.meta_id == 424
    assert hdr.schema_version == 1
    packed = pack(hdr.protocol_id, meta=SRMeta(hdr.schema_id, hdr.schema_version, hdr.meta_id))
    assert packed == serialized

    with pytest.raises(ValueError):
        pack(hdr.protocol_id, meta=SRMeta(hdr.schema_id, hdr.schema_version, None))


def test_handle_confluent() -> None:
    serialized = b'\x00\x00\x00\x00\x14'
    hdr = parse(serialized)
    assert hdr.protocol_id == 0
    assert hdr.meta_id is None
    assert hdr.schema_version is None
    packed = pack(hdr.protocol_id, meta=SRMeta(hdr.schema_id, hdr.schema_version, hdr.meta_id))
    assert packed == serialized


def test_handle_confluent_default_null() -> None:
    serialized = b'\x00\x00\x00\x00\x15'
    hdr = parse(serialized)
    assert hdr.protocol_id == 0
    assert hdr.meta_id is None
    assert hdr.schema_version is None
    packed = pack(hdr.protocol_id, meta=SRMeta(hdr.schema_id, hdr.schema_version, hdr.meta_id))
    assert packed == serialized



# ToDo (tribunsky.kir): looks like that idea of dynamic import via imp depending on librdkafka
#                       wasn't the worst idea, cause `if`s causes a lot of static checks problems.
# mypy: disable-error-code="no-redef,assignment"

from wunderkafka import librdkafka

# Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
#   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
#   and we don't want to have a nice whitelist, which is arguable
COMMON_FIELDS = (
    'api.version.fallback.ms',
    'api.version.request',
    'api.version.request.timeout.ms',
    'background_event_cb',
    'bootstrap.servers',
    'broker.address.family',
    'broker.address.ttl',
    'broker.version.fallback',
    'builtin.features',
    'client.id',
    'client.rack',
    'closesocket_cb',
    'connect_cb',
    'debug',
    'default_topic_conf',
    'enable.sasl.oauthbearer.unsecure.jwt',
    'enable.ssl.certificate.verification',
    'enabled_events',
    'error_cb',
    'interceptors',
    'internal.termination.signal',
    'log.connection.close',
    'log.queue',
    'log.thread.name',
    'log_cb',
    'log_level',
    'max.in.flight',
    'max.in.flight.requests.per.connection',
    'message.copy.max.bytes',
    'message.max.bytes',
    'metadata.broker.list',
    'metadata.max.age.ms',
    'metadata.request.timeout.ms',
    'oauthbearer_token_refresh_cb',
    'opaque',
    'open_cb',
    'plugin.library.paths',
    'receive.message.max.bytes',
    'reconnect.backoff.max.ms',
    'reconnect.backoff.ms',
    'sasl.kerberos.keytab',
    'sasl.kerberos.kinit.cmd',
    'sasl.kerberos.min.time.before.relogin',
    'sasl.kerberos.principal',
    'sasl.kerberos.service.name',
    'sasl.mechanism',
    'sasl.mechanisms',
    'sasl.oauthbearer.config',
    'sasl.password',
    'sasl.username',
    'security.protocol',
    'socket.keepalive.enable',
    'socket.max.fails',
    'socket.nagle.disable',
    'socket.receive.buffer.bytes',
    'socket.send.buffer.bytes',
    'socket.timeout.ms',
    'socket_cb',
    'ssl.ca.location',
    'ssl.certificate.location',
    'ssl.certificate.pem',
    'ssl.certificate.verify_cb',
    'ssl.cipher.suites',
    'ssl.crl.location',
    'ssl.curves.list',
    'ssl.endpoint.identification.algorithm',
    'ssl.key.location',
    'ssl.key.password',
    'ssl.key.pem',
    'ssl.keystore.location',
    'ssl.keystore.password',
    'ssl.sigalgs.list',
    'ssl_ca',
    'ssl_certificate',
    'ssl_key',
    'statistics.interval.ms',
    'stats_cb',
    'throttle_cb',
    'topic.blacklist',
    'topic.metadata.refresh.fast.interval.ms',
    'topic.metadata.refresh.interval.ms',
    'topic.metadata.refresh.sparse',
)
CONSUMER_FIELDS = (
    'auto.commit.interval.ms',
    'auto.offset.reset',
    'check.crcs',
    'consume.callback.max.messages',
    'consume_cb',
    'coordinator.query.interval.ms',
    'enable.auto.commit',
    'enable.auto.offset.store',
    'enable.partition.eof',
    'fetch.error.backoff.ms',
    'fetch.max.bytes',
    'fetch.message.max.bytes',
    'fetch.min.bytes',
    'fetch.wait.max.ms',
    'group.id',
    'group.instance.id',
    'group.protocol.type',
    'heartbeat.interval.ms',
    'isolation.level',
    'max.partition.fetch.bytes',
    'max.poll.interval.ms',
    'offset_commit_cb',
    'partition.assignment.strategy',
    'queued.max.messages.kbytes',
    'queued.min.messages',
    'rebalance_cb',
    'session.timeout.ms',
)
PRODUCER_FIELDS = (
    'acks',
    'batch.num.messages',
    'compression.codec',
    'compression.level',
    'compression.type',
    'delivery.report.only.error',
    'delivery.timeout.ms',
    'dr_cb',
    'dr_msg_cb',
    'enable.gapless.guarantee',
    'enable.idempotence',
    'linger.ms',
    'message.send.max.retries',
    'message.timeout.ms',
    'msg_order_cmp',
    'partitioner',
    'partitioner_cb',
    'queue.buffering.backpressure.threshold',
    'queue.buffering.max.kbytes',
    'queue.buffering.max.messages',
    'queue.buffering.max.ms',
    'queuing.strategy',
    'request.required.acks',
    'request.timeout.ms',
    'retries',
    'retry.backoff.ms',
    'transaction.timeout.ms',
    'transactional.id',
)

if librdkafka.__version__ >= (1, 5, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'metadata.request.timeout.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.config',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.location',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'allow.auto.create.topics',
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (1, 6, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'metadata.request.timeout.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.config',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'allow.auto.create.topics',
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (1, 7, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'connections.max.idle.ms',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.config',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.engine.id',
        'ssl.engine.location',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_engine_callback_data',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'allow.auto.create.topics',
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (1, 8, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'connections.max.idle.ms',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.config',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.engine.id',
        'ssl.engine.location',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_engine_callback_data',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'allow.auto.create.topics',
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (1, 9, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'connections.max.idle.ms',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.client.id',
        'sasl.oauthbearer.client.secret',
        'sasl.oauthbearer.config',
        'sasl.oauthbearer.extensions',
        'sasl.oauthbearer.method',
        'sasl.oauthbearer.scope',
        'sasl.oauthbearer.token.endpoint.url',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.connection.setup.timeout.ms',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.ca.pem',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.engine.id',
        'ssl.engine.location',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_engine_callback_data',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'allow.auto.create.topics',
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (2, 0, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'allow.auto.create.topics',
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'connections.max.idle.ms',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'resolve_cb',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.client.id',
        'sasl.oauthbearer.client.secret',
        'sasl.oauthbearer.config',
        'sasl.oauthbearer.extensions',
        'sasl.oauthbearer.method',
        'sasl.oauthbearer.scope',
        'sasl.oauthbearer.token.endpoint.url',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.connection.setup.timeout.ms',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.ca.pem',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.engine.id',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.providers',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_engine_callback_data',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (2, 1, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'allow.auto.create.topics',
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'connections.max.idle.ms',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'resolve_cb',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.client.id',
        'sasl.oauthbearer.client.secret',
        'sasl.oauthbearer.config',
        'sasl.oauthbearer.extensions',
        'sasl.oauthbearer.method',
        'sasl.oauthbearer.scope',
        'sasl.oauthbearer.token.endpoint.url',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.connection.setup.timeout.ms',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.ca.pem',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.engine.id',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.providers',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_engine_callback_data',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )

if librdkafka.__version__ >= (2, 2, 0):
    # Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',
    #   therefore, we can't convert on-the-fly from  `ssl_ca` without errors
    #   and we don't want to have a nice whitelist, which is arguable
    COMMON_FIELDS = (
        'allow.auto.create.topics',
        'api.version.fallback.ms',
        'api.version.request',
        'api.version.request.timeout.ms',
        'background_event_cb',
        'bootstrap.servers',
        'broker.address.family',
        'broker.address.ttl',
        'broker.version.fallback',
        'builtin.features',
        'client.dns.lookup',
        'client.id',
        'client.rack',
        'closesocket_cb',
        'connect_cb',
        'connections.max.idle.ms',
        'debug',
        'default_topic_conf',
        'enable.random.seed',
        'enable.sasl.oauthbearer.unsecure.jwt',
        'enable.ssl.certificate.verification',
        'enabled_events',
        'error_cb',
        'interceptors',
        'internal.termination.signal',
        'log.connection.close',
        'log.queue',
        'log.thread.name',
        'log_cb',
        'log_level',
        'max.in.flight',
        'max.in.flight.requests.per.connection',
        'message.copy.max.bytes',
        'message.max.bytes',
        'metadata.broker.list',
        'metadata.max.age.ms',
        'oauthbearer_token_refresh_cb',
        'opaque',
        'open_cb',
        'plugin.library.paths',
        'receive.message.max.bytes',
        'reconnect.backoff.max.ms',
        'reconnect.backoff.ms',
        'resolve_cb',
        'sasl.kerberos.keytab',
        'sasl.kerberos.kinit.cmd',
        'sasl.kerberos.min.time.before.relogin',
        'sasl.kerberos.principal',
        'sasl.kerberos.service.name',
        'sasl.mechanism',
        'sasl.mechanisms',
        'sasl.oauthbearer.client.id',
        'sasl.oauthbearer.client.secret',
        'sasl.oauthbearer.config',
        'sasl.oauthbearer.extensions',
        'sasl.oauthbearer.method',
        'sasl.oauthbearer.scope',
        'sasl.oauthbearer.token.endpoint.url',
        'sasl.password',
        'sasl.username',
        'security.protocol',
        'socket.connection.setup.timeout.ms',
        'socket.keepalive.enable',
        'socket.max.fails',
        'socket.nagle.disable',
        'socket.receive.buffer.bytes',
        'socket.send.buffer.bytes',
        'socket.timeout.ms',
        'socket_cb',
        'ssl.ca.certificate.stores',
        'ssl.ca.location',
        'ssl.ca.pem',
        'ssl.certificate.location',
        'ssl.certificate.pem',
        'ssl.certificate.verify_cb',
        'ssl.cipher.suites',
        'ssl.crl.location',
        'ssl.curves.list',
        'ssl.endpoint.identification.algorithm',
        'ssl.engine.id',
        'ssl.key.location',
        'ssl.key.password',
        'ssl.key.pem',
        'ssl.keystore.location',
        'ssl.keystore.password',
        'ssl.providers',
        'ssl.sigalgs.list',
        'ssl_ca',
        'ssl_certificate',
        'ssl_engine_callback_data',
        'ssl_key',
        'statistics.interval.ms',
        'stats_cb',
        'throttle_cb',
        'topic.blacklist',
        'topic.metadata.propagation.max.ms',
        'topic.metadata.refresh.fast.interval.ms',
        'topic.metadata.refresh.interval.ms',
        'topic.metadata.refresh.sparse',
    )
    CONSUMER_FIELDS = (
        'auto.commit.interval.ms',
        'auto.offset.reset',
        'check.crcs',
        'consume.callback.max.messages',
        'consume_cb',
        'coordinator.query.interval.ms',
        'enable.auto.commit',
        'enable.auto.offset.store',
        'enable.partition.eof',
        'fetch.error.backoff.ms',
        'fetch.max.bytes',
        'fetch.message.max.bytes',
        'fetch.min.bytes',
        'fetch.queue.backoff.ms',
        'fetch.wait.max.ms',
        'group.id',
        'group.instance.id',
        'group.protocol.type',
        'heartbeat.interval.ms',
        'isolation.level',
        'max.partition.fetch.bytes',
        'max.poll.interval.ms',
        'offset_commit_cb',
        'partition.assignment.strategy',
        'queued.max.messages.kbytes',
        'queued.min.messages',
        'rebalance_cb',
        'session.timeout.ms',
    )
    PRODUCER_FIELDS = (
        'acks',
        'batch.num.messages',
        'batch.size',
        'compression.codec',
        'compression.level',
        'compression.type',
        'delivery.report.only.error',
        'delivery.timeout.ms',
        'dr_cb',
        'dr_msg_cb',
        'enable.gapless.guarantee',
        'enable.idempotence',
        'linger.ms',
        'message.send.max.retries',
        'message.timeout.ms',
        'msg_order_cmp',
        'partitioner',
        'partitioner_cb',
        'queue.buffering.backpressure.threshold',
        'queue.buffering.max.kbytes',
        'queue.buffering.max.messages',
        'queue.buffering.max.ms',
        'queuing.strategy',
        'request.required.acks',
        'request.timeout.ms',
        'retries',
        'retry.backoff.ms',
        'sticky.partitioning.linger.ms',
        'transaction.timeout.ms',
        'transactional.id',
    )




from typing import Any, List, Union, Optional
from dataclasses import dataclass

from wunderkafka.types import DeliveryCallback
from wunderkafka.callbacks import error_callback
from wunderkafka.producers.bytes import BytesProducer


@dataclass
class Message(object):
    topic: str
    value: Optional[Union[str, bytes]]
    key: Optional[Union[str, bytes]]


class TestProducer(BytesProducer):

    def __init__(self) -> None:
        self.sent: List[Message] = []

    def send_message(
        self,
        topic: str,
        value: Optional[Union[str, bytes]] = None,
        key: Optional[Union[str, bytes]] = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = error_callback,
        *args: Any,
        blocking: bool = False,
        **kwargs: Any,
    ) -> None:
        message = Message(topic, value, key)
        self.sent.append(message)



import time
import datetime

import pytest

from wunderkafka.time import ts2dt


def test_invalid_ts() -> None:
    with pytest.raises(ValueError):
        ts2dt(time.time()*1000000)


def test_seconds() -> None:
    dt = datetime.datetime.now()
    assert ts2dt(dt.timestamp()) == dt


def test_ms() -> None:
    dt = datetime.datetime.now()
    assert ts2dt(dt.timestamp()*1000) - dt.replace(microsecond=0) <= datetime.timedelta(seconds=1)



import sys
import json
from typing import Optional

from pydantic import Field, BaseModel

from wunderkafka.serdes.avromodel import derive

if sys.version_info <= (3, 10):
    class ParentOptional(BaseModel):
        volume: float = Field(description='...')
        weight: Optional[float] = Field(description='...')

    class ChildOptional(ParentOptional):
        ts: int = Field(description='...')

    class GrandsonOptional(ChildOptional):
        ...


    def test_dataclass() -> None:
        schema = derive(GrandsonOptional, topic='test_data_1')
        assert json.loads(schema) == {
            'type': 'record',
            'name': 'test_data_1_value',
            'fields': [
                {
                    'name': 'volume',
                    'type': 'double',
                },
                {
                    'name': 'weight',
                    'type': ['double', 'null'],
                },
                {
                    'name': 'ts',
                    'type': 'long',
                },
            ],
        }

else:
    class Parent(BaseModel):
        volume: float = Field(description='...')
        weight: float | None = Field(description='...')


    class Child(Parent):
        ts: int = Field(description='...')


    class Grandson(Child):
        ...

    def test_dataclass_pipe_annotation() -> None:
        schema = derive(Grandson, topic='test_data_1')
        assert json.loads(schema) == {
            'type': 'record',
            'name': 'test_data_1_value',
            'fields': [
                {
                    'name': 'volume',
                    'type': 'double',
                },
                {
                    'name': 'weight',
                    'type': ['double', 'null'],
                },
                {
                    'name': 'ts',
                    'type': 'long',
                },
            ],
        }



"""
Module contains interface-like skeletons for consumer.

- inherited from confluent-kafka Consumer to use it as nested entity
- high-level consumer which is able to handle message's schemas
"""

import datetime
from abc import ABC, abstractmethod
from typing import Dict, List, Union, Optional

from confluent_kafka import Message, Consumer, TopicPartition

from wunderkafka.types import HowToSubscribe
from wunderkafka.config import ConsumerConfig
from wunderkafka.consumers.subscription import TopicSubscription


class AbstractConsumer(Consumer):
    """Extension point for the original Consumer API."""

    # Why so: https://github.com/python/mypy/issues/4125
    _config: ConsumerConfig
    subscription_offsets: Optional[Dict[str, HowToSubscribe]] = None

    @property
    def config(self) -> ConsumerConfig:
        """
        Get the consumer's config.

        Tech Debt: needed here for specific callback which resets partition to specific offset.

        :return:        Pydantic model with librdkafka consumer's configuration.
        """
        return self._config

    def batch_poll(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        raise_on_lost: bool = False,
    ) -> List[Message]:
        """
        Consume as many messages as we can for a given timeout.

        Created to allow deserializing consumer nest BytesConsumer and read messages via batches
        and not communicate with broker for every single message.

        :param timeout:         The maximum time to block waiting for message.
        :param num_messages:    The maximum number of messages to receive from broker.
                                Default is 1000000 which was the allowed maximum for librdkafka 1.2.
        :param raise_on_lost:   If True, raise exception whenever we are too late to call next poll.
                                Otherwise, will do nothing or behave depending on specified callbacks.

        :return:                A list of Message objects (possibly empty on timeout).

        ..  # noqa: DAR401
        ..  # noqa: DAR202
        """
        raise NotImplementedError


class AbstractDeserializingConsumer(ABC):
    """High-level interface for extended consumer."""

    @abstractmethod
    def commit(
        self,
        message: Optional[Message] = None,
        offsets: Optional[List[TopicPartition]] = None,
        asynchronous: bool = True,
    ) -> Optional[List[TopicPartition]]:
        """
        Commit a message or a list of offsets.

        This method overlaps original consumer's method and will use the nested consumer.

        :param message:         Commit offset (+1), extracted from Message object itself.
        :param offsets:         Commit exactly TopicPartition data.
        :param asynchronous:    If True, do not block execution, otherwise - wait until commit fail or success.

        :raises KafkaException: If all commits failed.

        :return:                On asynchronous call returns None immediately.
                                Committed offsets on synchronous call, if succeed.
        """

    @abstractmethod
    def subscribe(  # noqa: WPS211  # ToDo (tribunsky.kir): reconsider API of 'how'
        self,
        topics: List[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        """
        Subscribe to a given list of topics. This replaces a previous subscription.

        This method overlaps original consumer's method and will use the nested consumer.

        :param topics:          List of topics to subscribe. If topic has no specific subscription, specified value
                                for beginning/end/offset/timestamp/timedelta/built-in will be used.
        :param from_beginning:  If flag is set, return specific offset corresponding to beginning/end of the topic.
        :param offset:          If set, will return Offset object.
        :param ts:              If set, will return Timestamp.
        :param with_timedelta:  If set, will calculate timestamp for corresponding timedelta.
        """



import json
import time
from typing import Optional
from datetime import datetime
from dataclasses import dataclass

import pytest
from pydantic import Field, BaseModel, ValidationError
from pydantic_settings import BaseSettings

from dataclasses_avroschema import AvroModel
from wunderkafka.serdes.avromodel import derive

# ToDo (tribunsky.kir): review some tests. As pydantic V2 changes it's behaviour, some tests are useless:
#                       we still can derive correct model, but we will be unable to actually populate it in runtime.
# ToDo (tribunsky.kir): switch to derivation of dataclasses-avroschema[pydantic], when it start support v2:
#                       https://github.com/marcosschroh/dataclasses-avroschema/issues/381


@dataclass
class SomeData(AvroModel):
    field1: int
    field2: str


@dataclass
class SomeDefaultData(AvroModel):
    field1: Optional[int] = None
    field2: Optional[str] = None


class Metrics(BaseModel):
    line_speed: Optional[int]
    defect_detected: Optional[bool] = False
    model_on: Optional[bool] = False
    squad_number: int = 0


class Metric(BaseSettings):
    line_speed: Optional[int]
    defect_detected: Optional[bool] = False
    model_on: Optional[bool] = False
    squad_number: int = 0


class MetricV2(BaseSettings):
    line_speed: Optional[int]
    defect_detected: Optional[bool] = False
    model_on: Optional[bool] = False
    squad_number: int = 0
    model_config: str                                                                                     # type: ignore


class Event(BaseModel):
    id: Optional[int]
    ts: Optional[int] = None

    class Meta:
        namespace = "any.data"


class Mixed(BaseModel):
    text: str
    value: float
    integer: int = 0
    string: str = 'str'


# Pydantic allows following of non-default arguments, but dataclasses are not.
class ImageData(BaseModel):
    name: str = 'str'
    image: bytes
    camera: str
    ts: float = Field(default_factory=time.time)


class OtherImageData(BaseModel):
    image: bytes
    camera: str
    ts: float = Field(default_factory=time.time)
    name: str = 'str'


class TsWithMeta(BaseModel):
    ts: datetime = Field(default_factory=datetime.now)

    class Meta:
        namespace = 'com.namespace.my'
        name = 'MsgKey'


def test_dataclass() -> None:
    schema = derive(SomeData, topic='test_data_1')
    assert json.loads(schema) == {
        'type': 'record',
        'name': 'test_data_1_value',
        'fields': [
            {
                'name': 'field1',
                'type': 'long',
            },
            {
                'name': 'field2',
                'type': 'string',
            },
        ],
    }


def test_dataclass_defaults() -> None:
    schema = derive(SomeDefaultData, topic='test_data_2')

    assert json.loads(schema) == {
        'type': 'record',
        'name': 'test_data_2_value',
        'fields': [
            {
                'name': 'field1',
                'type': ['null', 'long'],
                'default': None,
            },
            {
                'name': 'field2',
                'type': ['null', 'string'],
                'default': None,
            },
        ],
    }


def test_pydantic_with_defaults() -> None:
    schema = derive(Metrics, topic='some_topic')

    assert json.loads(schema) == {
      "type": "record",
      "name": "some_topic_value",
      "fields": [
        {
          "name": "line_speed",
          "type": [
            "long",
            "null"
          ]
        },
        {
          "name": "defect_detected",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "model_on",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "squad_number",
          "type": "long",
          "default": 0,
        }
      ]
    }


def test_pydantic_defaults() -> None:
    schema = derive(Mixed, topic='topic')

    assert json.loads(schema) == {
        'type': 'record',
        'name': 'topic_value',
        'fields': [
            {
                'type': 'string',
                'name': 'text',
            },
            {
                'type': 'double',
                'name': 'value',
            },
            {
                'type': 'long',
                'name': 'integer',
                'default': 0,
            },
            {
                'type': 'string',
                'name': 'string',
                'default': 'str'
            },
        ],
    }


def test_pydantic_defaults_none() -> None:
    schema = derive(Event, topic='topic')

    assert json.loads(schema) == {
        'type': 'record',
        'name': 'topic_value',
        'namespace': 'any.data',
        'fields': [
            {
                'type': ['long', 'null'],
                'name': 'id',
            },
            {
                'type': ['null', 'long'],
                'name': 'ts',
                'default': None,
            },
        ],
    }


def test_pydantic_mixed_defaults() -> None:
    s1 = derive(ImageData, topic='topic')
    s2 = derive(OtherImageData, topic='topic')
    assert json.loads(s1) == {
        'type': 'record',
        'name': 'topic_value',
        'fields': [
            {
                'name': 'name',
                'type': 'string',
                'default': 'str'
            },
            {
                'name': 'image',
                'type': 'bytes',
            },
            {
                'name': 'camera',
                'type': 'string',
            },
            {
                'name': 'ts',
                'type': 'double',
            },
        ],
    }
    assert json.loads(s1) != json.loads(s2)


def test_pydantic_with_meta() -> None:
    schema = derive(TsWithMeta, topic='topic')

    assert json.loads(schema) == {
        'type': 'record',
        'namespace': 'com.namespace.my',
        'name': 'MsgKey',
        'fields': [
            {
                'name': 'ts',
                'type': {
                    'logicalType': 'timestamp-millis',
                    'type': 'long',
                }
            }
        ],
    }


def test_pydantic_base_settings_with_defaults() -> None:
    schema = derive(Metric, topic='some_topic')

    assert json.loads(schema) == {
      "type": "record",
      "name": "some_topic_value",
      "fields": [
        {
          "name": "line_speed",
          "type": [
            "long",
            "null"
          ]
        },
        {
          "name": "defect_detected",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "model_on",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "squad_number",
          "type": "long",
          "default": 0,
        }
      ]
    }


def test_pydantic_base_settings_v2_with_defaults() -> None:
    schema = derive(MetricV2, topic='some_topic')

    assert json.loads(schema) == {
      "type": "record",
      "name": "some_topic_value",
      "fields": [
        {
          "name": "line_speed",
          "type": [
            "long",
            "null"
          ]
        },
        {
          "name": "defect_detected",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "model_on",
          "type": [
            "boolean",
            "null"
          ],
          "default": False
        },
        {
          "name": "squad_number",
          "type": "long",
          "default": 0,
        },
        {
          "name": "model_config",
          "type": "string",
        }
      ]
    }

    # model_config is a ClassVar, so we can't populate model anyway
    with pytest.raises(ValidationError):
        MetricV2(line_speed=2, model_config='str')                                                        # type: ignore



from wunderkafka.schema_registry.cache import SimpleCache
from wunderkafka.schema_registry.transport import KerberizableHTTPClient
from wunderkafka.schema_registry.clients.cloudera import ClouderaSRClient
from wunderkafka.schema_registry.clients.confluent import ConfluentSRClient



from unittest.mock import Mock

from pytest import MonkeyPatch

from wunderkafka.hotfixes.watchdog import KrbWatchDog, KinitParams


def test_krb_watchdog(monkeypatch: MonkeyPatch) -> None:
    params = [KinitParams(user=f'user_{i}', realm='realm', keytab=f'keytab{i}', cmd='cmd') for i in range(10)]
    monkeypatch.setattr('wunderkafka.hotfixes.watchdog.KrbWatchDog._ensure_started', Mock())
    for i, p in enumerate(params, start=1):
        thread = KrbWatchDog()
        thread.add(p)
        assert thread.count_params == i



from typing import Any, Dict

from confluent_kafka import KafkaError

from wunderkafka.config.generated import enums
from wunderkafka.config.rdkafka import RDKafkaConfig
from wunderkafka.logger import logger


def exclude_gssapi(builtin_features: str) -> str:
    features = [feature.strip() for feature in builtin_features.split(',') if feature.strip() != 'sasl_gssapi']
    return ', '.join(features)


def config_requires_kerberos(config: RDKafkaConfig) -> bool:
    if config.sasl_mechanism.casefold() != 'GSSAPI'.casefold():
        return False
    if config.sasl_kerberos_keytab is None:
        return False
    return config.security_protocol in {enums.SecurityProtocol.sasl_ssl, enums.SecurityProtocol.sasl_plaintext}


def challenge_krb_arg(exc: KafkaError, config: RDKafkaConfig) -> RDKafkaConfig:
    """
    Check if we can just skip kerberos configuration which comes to RDKafkaConfig from documentation default.

    Currently, I didn't find anything in confluent-kafka to query `builtin.features` from python,
    so we are just checking error while instantiating original consumer/producer
    and override corresponding config values.
    """
    logger.warning('Error while instantiating consumer/producer. Checking builtin.features...')
    # https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafkaexception
    error = exc.args[0]
    if error.code() != KafkaError._INVALID_ARG:
        raise
    if 'sasl_gssapi' not in error.str():
        raise
    if config_requires_kerberos(config):
        raise
    else:
        logger.warning(' '.join([
            "Looks like that current client configuration doesn't require kerberos.",
            "As it is not supported by currently installed build, skipping this option.",
        ]))
        old = config.builtin_features
        new = exclude_gssapi(config.builtin_features)
        logger.warning('Changing builtin.features: {0} -> {1}'.format(old, new))
        config.builtin_features = new
        return config



import os
import json
from typing import Any

import requests

from wunderkafka.config.krb.schema_registry import HTTPKerberosAuth
from wunderkafka.logger import logger
from wunderkafka.schema_registry.abc import AbstractHTTPClient
from wunderkafka.config.schema_registry import SRConfig


class KerberizableHTTPClient(AbstractHTTPClient):
    def __init__(self, config: SRConfig, *, requires_kerberos: bool = False, save_replay: bool = False) -> None:
        s = requests.Session()
        if requires_kerberos:
            s.auth = HTTPKerberosAuth(
                principal=config.sasl_username,
                mutual_authentication=config.mutual_auth,
            )

        if config.ssl_certificate_location is not None:
            if config.ssl_key_location is not None:
                s.cert = (config.ssl_certificate_location, config.ssl_key_location)
            else:
                s.cert = config.ssl_certificate_location

        if config.ssl_key_location is not None:
            s.verify = config.ssl_ca_location

        accept = ', '.join([
            'application/vnd.schemaregistry.v1+json',
            'application/vnd.schemaregistry+json',
            'application/json',
        ])
        s.headers.update({
            'Accept': accept,
            'Content-Type': "application/vnd.schemaregistry.v1+json",
        })

        self._session = s
        self._base_url = config.url

        self._save_replay = save_replay

    @property
    def base_url(self) -> str:
        return str(self._base_url)

    def close(self) -> None:
        self._session.close()

    # ToDo (tribunsky.kir): make method enum
    def make_request(
        self,
        relative_url: str,
        method: str = 'GET',
        body: Any = None,
        query: Any = None,
    ) -> Any:
        url = "/".join([self._base_url.rstrip('/'), relative_url.lstrip('/')])
        logger.debug('{0}: {1}'.format(method, url))
        response = self._session.request(method, url, json=body, params=query)
        # ToDo(aa.perelygin): more informative message when fail
        if response.status_code >= 400:
            logger.error(f'HTTPError for url: {url}, more details: {response.content}')

        response.raise_for_status()

        self._dump(method, relative_url, response)
        return response.json()

    def _dump(self, method: str, relative_url: str, response: requests.Response) -> None:
        if self._save_replay:
            filename = '{0}/{1}.json'.format(method, relative_url)
            dir_name = os.path.dirname(filename)
            os.makedirs(dir_name, exist_ok=True)
            with open(filename, 'w') as fl:
                json.dump(response.json(), fl)



import json
from typing import Any, Dict, Type
from dataclasses import is_dataclass

from pydantic import BaseModel

from dataclasses_avroschema import AvroModel
from pydantic_core import PydanticUndefinedType


def derive(model: Type[object], topic: str, *, is_key: bool = False) -> str:
    if is_dataclass(model):
        # https://github.com/python/mypy/issues/14941
        model_schema = model.avro_schema_to_python()                                                      # type: ignore
    else:
        # non-dataclasses objects may allow mixing defaults and non-default fields order,
        # so to still reuse dataclasses_avroschema, we extract fields and reorder them to satisfy dataclasses
        # restrictions, than reorder them back.
        # All this hacks will work only for flat schemas: we avoid describing objects via nested types for HDFS's sake.
        attributes = _extract_attributes(model)
        ordering = list(attributes['__annotations__'])
        crafted_model = _construct_model(attributes, model)
        model_schema = crafted_model.avro_schema_to_python()
        fields_map = {field_data['name']: field_data for field_data in model_schema['fields']}
        reordered_fields = [fields_map[attr] for attr in ordering]
        model_schema['fields'] = reordered_fields
    model_schema.pop('doc', None)
    suffix = 'key' if is_key else 'value'
    if hasattr(model, 'Meta') and hasattr(model.Meta, 'name'):
        model_schema['name'] = model.Meta.name
    else:
        model_schema['name'] = '{0}_{1}'.format(topic, suffix)
    return json.dumps(model_schema)


def _construct_model(attrs: Dict[str, Any], type_: Type[object]) -> AvroModel:
    if issubclass(type_, BaseModel):
        for field_name, field_info in type_.model_fields.items():
            if not isinstance(field_info.default, PydanticUndefinedType):
                attrs[field_name] = field_info.default
                tp = attrs['__annotations__'].pop(field_name)
                attrs['__annotations__'].update({field_name: tp})
    # https://docs.python.org/3/library/functions.html?highlight=type#type
    return type(type_.__name__, (AvroModel,), attrs)                                                      # type: ignore


def _extract_attributes(type_: Type[object]) -> Dict[str, Any]:
    fields = vars(type_).get('__annotations__', {})
    _, *parents = type_.mro()
    for base in parents:
        new_fields = {**vars(base).get('__annotations__', {})}
        # Currently it is `model_config: ClassVar[SettingsConfigDict]`
        # https://github.com/pydantic/pydantic-settings/blob/919a20b77527ecc1cd6eeb0a09ca22cc21486fb8/pydantic_settings/main.py#L166
        for field in ['__slots__', 'klass', 'metadata', 'schema_def', '__config__', 'model_config']:
            new_fields.pop(field, None)
        fields = {**new_fields, **fields}

    attributes = {}
    attributes.update({'__annotations__': fields})
    meta = vars(type_).get('Meta', None)
    if meta is not None:
        attributes.update({'Meta': meta})
    return attributes



import os
import time
import subprocess
from typing import Any, Set, Dict, Tuple, Optional
from threading import Thread
from dataclasses import dataclass

from confluent_kafka import libversion

from wunderkafka.time import ts2dt
from wunderkafka.logger import logger
from wunderkafka.config.rdkafka import RDKafkaConfig
from wunderkafka.config.generated import enums
from wunderkafka.hotfixes.watchdog.types import Watchdog
from wunderkafka.hotfixes.watchdog.krb.ticket import get_expiration_ts

REQUIRES_KERBEROS = frozenset([enums.SecurityProtocol.sasl_ssl, enums.SecurityProtocol.sasl_plaintext])


class NonRepetitiveLogger(object):

    def __init__(self) -> None:
        self._logged: Set[int] = set()

    def _log(self, message: str, *, info: bool = True) -> None:
        hashed = hash(message)
        if hashed in self._logged:
            return
        if info:
            logger.info(message)
        else:
            logger.warning(message)
        self._logged.add(hashed)

    def info(self, message: str) -> None:
        self._log(message)

    def warning(self, message: str) -> None:
        self._log(message, info=False)


log_once = NonRepetitiveLogger()


@dataclass(frozen=True)
class KinitParams(object):
    user: str
    realm: str
    keytab: str
    cmd: str

    @property
    def keytab_filename(self) -> str:
        return self.keytab.split(os.path.sep)[-1]

    @property
    def principal(self) -> str:
        return '{0}@{1}'.format(self.user, self.realm)


def get_version() -> Tuple[int, ...]:
    semver, _ = libversion()
    return split(semver)


def split(version: str) -> Tuple[int, ...]:
    return tuple(int(digit) for digit in version.split('.')[:2])


# ToDo (tribunsky.kir): Currently will log for every created producer/consumer. Move messaging to Watchdog itself.
def check_watchdog(
    config: RDKafkaConfig,
    version: Tuple[int, ...] = get_version(),
) -> Tuple[RDKafkaConfig, Optional[Watchdog]]:
    watchdog = None
    if config.security_protocol.value not in REQUIRES_KERBEROS:
        return config, watchdog
    if version > (1, 7):
        log_once.info('Watchdog: please check if there are any issues with SASL GSSAPI in the upstream.')
        return config, watchdog

    if version == (1, 7):
        msg = 'rewriting config. For more info, please, check https://github.com/edenhill/librdkafka/issues/3430'
        log_once.warning('Watchdog: {0}'.format(msg))
        config.sasl_kerberos_min_time_before_relogin = 86400000
    elif version < (1, 7):
        msg = ' '.join([
            'adding watchdog thread.',
            'Prior to 1.7.0 kinit refresh via librdkafka may cause blocking.',
            'For more info, please, check https://github.com/edenhill/librdkafka/pull/3340',
        ])
        log_once.warning('Watchdog: {0}'.format(msg))
        params = parse_kinit(config.sasl_kerberos_kinit_cmd)
        config.sasl_kerberos_min_time_before_relogin = 0
        watchdog = KrbWatchDog()
        watchdog.add(params)
    return config, watchdog


class Borg(object):
    _shared_state: Dict[str, Any] = {}

    def __init__(self) -> None:
        self.__dict__ = self._shared_state


def parse_kinit(kinit_cmd: str) -> KinitParams:
    kinit_cmd_msg = 'kinit_cmd: `{0}`'.format(kinit_cmd)
    parts = []
    keytab = None
    prev = None
    for word in kinit_cmd.split():
        normalized = word.strip()
        if prev == '-t':
            keytab = normalized
        elif len(normalized) > 2 and normalized != 'kinit':
            parts.append(normalized)
        prev = normalized
    if keytab is None:
        raise ValueError("Couldn't get keytab: {0} ({1})".format(keytab, kinit_cmd_msg))
    if len(parts) != 1:
        raise ValueError("Couldn't parse {0}".format(kinit_cmd_msg))
    [principal] = parts
    delim = '@'
    user, realm = principal.split(delim)
    if not (user and realm):
        raise ValueError("Couldn't parse principal: {0} ({1})".format(principal, kinit_cmd_msg))
    return KinitParams(user=user, realm=realm, keytab=keytab, cmd=kinit_cmd)


class KerberosRefreshThread(Thread):
    def __init__(self, params: Set[KinitParams], default_timeout: int) -> None:
        super().__init__()
        self.daemon = True

        self._params_refresh: Dict[KinitParams, float] = {}
        for p in params:
            self._refresh_krb(p)

        self._default_timeout = default_timeout
        logger.info("Kerberos thread: initiated")

    @property
    def krb_timeout(self) -> int:
        return self._default_timeout

    @krb_timeout.setter
    def krb_timeout(self, value: int) -> None:
        self._default_timeout = value

    def run(self) -> None:
        logger.info("Kerberos thread: started")
        while True:
            for param in self._params_refresh:
                if time.time() >= self._params_refresh[param]:
                    self._refresh_krb(param)
            time.sleep(0.01)

    def update_keytabs(self, params: Set[KinitParams]) -> None:
        for param in params:
            if param not in self._params_refresh:
                self._refresh_krb(param)
                logger.debug("The addition of the new keytab ({0}|{1}) was successful".format(
                    param.principal, param.keytab_filename
                ))

    def _refresh_krb(self, params: KinitParams) -> None:
        t0 = time.perf_counter()
        refresh_cmd = params.cmd.split()
        logger.info('Refreshing krb-ticket ({0}|{1})...'.format(params.principal, params.keytab_filename))
        try:
            subprocess.run(refresh_cmd, timeout=self._default_timeout, stdout=subprocess.PIPE, check=True)
        # Will retry shortly
        except (subprocess.CalledProcessError, subprocess.TimeoutExpired) as exc:
            logger.error(exc.output)
            logger.error(exc.stdout)
            logger.error(exc.stderr)
            logger.error('Command: {0} exit error: {1}'.format(refresh_cmd, str(exc)))
            logger.warning("Krb not refreshed!")
        else:
            duration = int(1000 * (time.perf_counter() - t0))
            logger.info('Refreshed! ({0} ms)'.format(duration))
            self._params_refresh[params] = get_expiration_ts(params.user, params.realm, self._default_timeout)
            logger.debug('Nearest existing: {0} keytab: ({1}|{2})'.format(
                ts2dt(self._params_refresh[params]), params.principal, params.keytab_filename)
            )


class KrbWatchDog(Borg):
    __thread: Optional[KerberosRefreshThread] = None
    __kinit_params: Set[KinitParams] = set()
    __default_timeout: int = 60

    def __init__(self) -> None:
        super().__init__()
        self._ensure_started()

    def add(self, params: KinitParams) -> None:
        if params not in self.__kinit_params:
            self.__kinit_params.add(params)
            self._ensure_started()

    @property
    def krb_timeout(self) -> int:
        return self.__default_timeout

    @krb_timeout.setter
    def krb_timeout(self, value: int) -> None:
        if self.__thread is not None:
            if int(value) > 0:
                self.__default_timeout = value
            self.__thread.krb_timeout = self.__default_timeout

    @property
    def count_params(self) -> int:
        return len(self.__kinit_params)

    def _ensure_started(self) -> None:
        if self.__thread is None:
            self.__thread = KerberosRefreshThread(self.__kinit_params, self.__default_timeout)
            self.__thread.start()
        else:
            self.__thread.update_keytabs(self.__kinit_params)
        if self.__thread.is_alive() is False:
            self.__thread = KerberosRefreshThread(self.__kinit_params, self.__default_timeout)
            self.__thread.start()

    def __call__(self) -> None:
        self._ensure_started()



from abc import ABC, abstractmethod
from typing import Any, Dict, Union, Optional

from wunderkafka.types import TopicName, KeySchemaDescription, ValueSchemaDescription
from wunderkafka.structures import SRMeta, ParsedHeader


class AbstractProtocolHandler(ABC):

    @abstractmethod
    def parse(self, blob: bytes) -> ParsedHeader: ...

    # ToDo (tribunsky.kir): make it symmetrical and more general
    @abstractmethod
    def pack(self, protocol_id: int, meta: SRMeta) -> bytes: ...


# ToDo (tribunsky.kir): make it parametrized generic?
class AbstractDeserializer(ABC):

    @abstractmethod
    def deserialize(self, schema: str, payload: bytes, seek_pos: Optional[int] = None) -> Any: ...


class AbstractSerializer(ABC):
    @abstractmethod
    def serialize(self, schema: str, payload: Any, header: Optional[bytes] = None) -> bytes: ...


class AbstractDescriptionStore(ABC):

    def __init__(self) -> None:
        self._values: Dict[TopicName, ValueSchemaDescription] = {}
        self._keys: Dict[TopicName, KeySchemaDescription] = {}

    def get(
        self,
        topic: TopicName,
        *,
        is_key: bool = False,
    ) -> Optional[Union[ValueSchemaDescription, KeySchemaDescription]]:
        if is_key:
            return self._keys.get(topic)
        else:
            return self._values.get(topic)

    @abstractmethod
    def add(self, topic: TopicName, value: Any, key: Any) -> None: ...



from abc import ABC, abstractmethod
from typing import Any

from wunderkafka.structures import SRMeta, SchemaMeta


class AbstractHTTPClient(ABC):

    @abstractmethod
    def make_request(self, relative_url: str, method: str = 'GET', body: Any = None, query: Any = None) -> Any: ...

    @property
    def base_url(self) -> str:
        raise NotImplementedError

    @abstractmethod
    def close(self) -> None: ...


class AbstractSchemaRegistry(ABC):

    @abstractmethod
    def get_schema_text(self, schema_meta: SchemaMeta) -> str: ...

    @abstractmethod
    def register_schema(self, topic: str, schema_text: str, *, is_key: bool = True) -> SRMeta: ...



import datetime
from typing import Any, List, Union, Optional

from confluent_kafka import Message, TopicPartition
from typing_extensions import Protocol

# ToDo (tribunsky.kir): subject to change. It's not obvious how to merge together
#                       python-kafka/confluent-kafka and out own API, so currently
#                       it's just API of (de)serializing producer/consumer with the nested 'real' producer/consumer.
from wunderkafka.types import MsgKey, MsgValue, DeliveryCallback
from wunderkafka.consumers.subscription import TopicSubscription


class AnyConsumer(Protocol):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def commit(
        self,
        message: Optional[Message] = None,
        offsets: Optional[List[TopicPartition]] = None,
        asynchronous: bool = True,
    ) -> Optional[List[TopicPartition]]:
        """Just overlap nested 'real' consumer's offset."""

    def subscribe(
        self,
        topics: List[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        """Subscribe to a given list of topics. This replaces a previous subscription."""
        ...

    def consume(
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        ignore_keys: bool = False,
    ) -> List[Message]:
        """Consume as many messages as we can for a given timeout and decode them."""


class AnyProducer(Protocol):

    def __init__(self, *args: Any, **kwargs: Any) -> None:
        ...

    def send_message(
            self,
            topic: str,
            value: MsgValue = None,
            key: MsgKey = None,
            partition: Optional[int] = None,
            on_delivery: Optional[DeliveryCallback] = None,
            *args: Any,
            blocking: bool = False,
            **kwargs: Any,
    ) -> None:
        """Send encoded message to Kafka almost immediately."""

    def set_target_topic(self, topic: str, value: Any, key: Any = None, *, lazy: bool = False) -> None:
        """Make producer aware how it should work with specific topic."""

    def flush(self, timeout: Optional[float] = None) -> int:
        """Just overlap nested 'real' producer's flush."""



from __future__ import annotations

from typing import Optional, Any
from typing_extensions import Protocol
from confluent_kafka.schema_registry import SchemaRegistryClient as ConfluentSchemaRegistryClient, Schema

from wunderkafka.schema_registry import SimpleCache
from wunderkafka.structures import SRMeta, SchemaMeta
from wunderkafka.schema_registry.abc import AbstractHTTPClient, AbstractSchemaRegistry


class ConfluentRestClient(Protocol):

    def get(self, url: str, query: Optional[dict] = None) -> Any:
        ...

    def post(self, url: str, body: Optional[str], **kwargs: Any) -> Any:
        ...

    def delete(self, url: str) -> Any:
        ...

    def put(self, url: str, body: Optional[str] = None) -> Any:
        ...

    def send_request(self, url: str, method: str, body: Optional[str] = None, query: Optional[dict] = None) -> Any:
        ...

    def _close(self) -> None:
        ...


class Adapter(object):
    def __init__(self, http_client: AbstractHTTPClient) -> None:
        self._client = http_client

    def get(self, url: str, query: Optional[dict] = None) -> Any:
        return self._client.make_request(url, query=query)

    def post(self, url: str, body: Optional[str], **_: Any) -> Any:
        return self._client.make_request(url, method='POST', body=body)

    def delete(self, url: str) -> Any:
        return self._client.make_request(url, method='DELETE')

    def put(self, url: str, body: Optional[str] = None) -> Any:
        return self._client.make_request(url, method='PUT', body=body)

    def send_request(self, url: str, method: str, body: Optional[str] = None, query: Optional[dict] = None) -> Any:
        return self._client.make_request(url, method, body=body, query=query)

    def _close(self) -> None:
        return self._client.close()


class SchemaRegistryClient(ConfluentSchemaRegistryClient):

    @classmethod
    def from_client(cls, http_client: AbstractHTTPClient) -> SchemaRegistryClient:
        # Minimal initialization as we will override client with our own
        client = cls({'url': http_client.base_url})
        client._rest_client = Adapter(http_client)
        return client


class ConfluentSRClient(AbstractSchemaRegistry):

    def __init__(self, http_client: AbstractHTTPClient, _: Optional[SimpleCache] = None) -> None:
        self._client = SchemaRegistryClient.from_client(http_client)

    def get_schema_text(self, meta: SchemaMeta) -> str:
        schema = self._client.get_schema(meta.header.schema_id)
        return schema.schema_str

    def register_schema(self, topic: str, schema_text: str, *, is_key: bool = True) -> SRMeta:
        # FixMe (tribunsky.kir): lack of symmetry here - SchemaMeta knows about different vendors, but not vice versa.
        subject = '{0}{1}'.format(topic, '_key' if is_key else '_value')
        schema_id = self._client.register_schema(subject, Schema(schema_text, schema_type='AVRO'))
        return SRMeta(schema_id, schema_version=None)



import os
from enum import Enum
from types import MappingProxyType
from typing import Any, Dict, Union, Mapping, TypeVar, Optional

from wunderkafka.config.schema_registry import SRConfig
from wunderkafka.config.generated.fields import COMMON_FIELDS, CONSUMER_FIELDS, PRODUCER_FIELDS
from wunderkafka.config.generated.models import RDConsumerConfig, RDProducerConfig
from wunderkafka.logger import logger

CONF_CONSUMER_FIELDS = MappingProxyType({
    field_name.replace('.', '_'): field_name for field_name in (*COMMON_FIELDS, *CONSUMER_FIELDS)
})

CONF_PRODUCER_FIELDS = MappingProxyType({
    field_name.replace('.', '_'): field_name for field_name in (*COMMON_FIELDS, *PRODUCER_FIELDS)
})

ConfigValues = Union[str, int, bool, float]


def remap_properties(
    dct: Dict[str, Optional[ConfigValues]],
    mapping: Mapping[str, str],
) -> Dict[str, ConfigValues]:
    new_dct = {}
    for f_name, f_value in dct.items():
        if f_value is not None:
            to_add = f_value
            if isinstance(f_value, Enum):
                to_add = f_value.value
            new_dct[mapping[f_name]] = to_add
    return new_dct


# ToDo (tribunsky.kir): I need separate common place to do somthing with config before feeding it to librdkafka
#                       or write more complex generator and class hierarchy (e.g. platform-specific base classes)
#                       to not monkeypatch config before actually feeding it to librdkafka.
#                       #TypeSafety!!1
def sanitize(dct: Dict[str, ConfigValues]) -> Dict[str, ConfigValues]:
    # cimpl.KafkaException: KafkaError{
    #   ...
    #   "Configuration property "ssl.ca.certificate.stores" not supported in this build: configuration only valid on Windows"
    #   }
    if os.name != 'nt':
        property_name = 'ssl.ca.certificate.stores'
        property_value = dct.pop(property_name, None)
        if property_value is not None:
            logger.warning('Excluding {0}={1} as windows-only even it was set to default'.format(property_name, property_value))
    return dct


# I don't like mixing SR config and librdkafka config,
# but it's more handful for producer w/o schema (no need to nest config for librdkafka)
class ConsumerConfig(RDConsumerConfig):
    # https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
    sr: Optional[SRConfig] = None

    def dict(self, **kwargs: Any) -> Dict[str, ConfigValues]:
        dct = super().model_dump(**kwargs)
        dct.pop('sr')
        return sanitize(remap_properties(dct, CONF_CONSUMER_FIELDS))


class ProducerConfig(RDProducerConfig):
    # https://docs.pydantic.dev/latest/migration/#required-optional-and-nullable-fields
    sr: Optional[SRConfig] = None

    def dict(self, **kwargs: Any) -> Dict[str, ConfigValues]:
        dct = super().model_dump(**kwargs)
        dct.pop('sr')
        return sanitize(remap_properties(dct, CONF_PRODUCER_FIELDS))


RDKafkaConfig = TypeVar('RDKafkaConfig', ConsumerConfig, ProducerConfig)



from typing import Type, Optional
from pathlib import Path

from pydantic import BaseModel

from wunderkafka.tests import TestProducer, TestHTTPClient
from wunderkafka.serdes.avro import AvroModelSerializer, ConfluentClouderaHeadersHandler
from wunderkafka.serdes.store import AvroModelRepo
from wunderkafka.schema_registry import SimpleCache, ConfluentSRClient
from wunderkafka.producers.constructor import HighLevelSerializingProducer


class Event(BaseModel):
    id: Optional[int]
    ts: Optional[int]

    class Meta:
        namespace = "any.data"


class EvolvedEvent(Event):
    description: Optional[str] = None
    info: Optional[str] = 'test'


def test_avro_producer_create_schema(sr_root_create: Path, topic: str, schema_description: Type[Event] = Event) -> None:
    fixed_ts = 1632128298534
    test_producer = TestProducer()
    producer = HighLevelSerializingProducer(
        producer=test_producer,
        schema_registry=ConfluentSRClient(TestHTTPClient(sr_root_create), SimpleCache()),
        header_packer=ConfluentClouderaHeadersHandler().pack,
        serializer=AvroModelSerializer(),
        store=AvroModelRepo(),
        mapping={topic: schema_description},
        protocol_id=0,
    )

    key = None
    value = schema_description(id=None, ts=fixed_ts)

    producer.send_message(topic, value, key)

    [message] = test_producer.sent

    assert message.key is None
    assert message.value == b'\x00\x00\x00\x00\x14\x02\x00\xcc\xb8\xeb\xa6\x80_'


def test_avro_producer_existing_schema(
    sr_root_existing: Path,
    topic: str,
    schema_description: Type[Event] = Event,
) -> None:
    fixed_ts = 1632128298534
    test_producer = TestProducer()
    producer = HighLevelSerializingProducer(
        producer=test_producer,
        schema_registry=ConfluentSRClient(TestHTTPClient(sr_root_existing), SimpleCache()),
        header_packer=ConfluentClouderaHeadersHandler().pack,
        serializer=AvroModelSerializer(),
        store=AvroModelRepo(),
        mapping={topic: schema_description},
        protocol_id=0,
    )

    key = None
    value = schema_description(id=None, ts=fixed_ts)

    producer.send_message(topic, value, key)

    [message] = test_producer.sent

    assert message.key is None
    assert message.value == b'\x00\x00\x00\x00\x14\x02\x00\xcc\xb8\xeb\xa6\x80_'


def test_avro_producer_update_schema(
    sr_root_update: Path,
    topic: str,
    schema_description: Type[EvolvedEvent] = EvolvedEvent,
) -> None:
    fixed_ts = 1632128298534
    test_producer = TestProducer()
    producer = HighLevelSerializingProducer(
        producer=test_producer,
        schema_registry=ConfluentSRClient(TestHTTPClient(sr_root_update), SimpleCache()),
        header_packer=ConfluentClouderaHeadersHandler().pack,
        serializer=AvroModelSerializer(),
        store=AvroModelRepo(),
        mapping={topic: schema_description},
        protocol_id=0,
    )

    key = None
    value = schema_description(id=None, ts=fixed_ts)

    producer.send_message(topic, value, key)

    [message] = test_producer.sent

    assert message.key is None
    assert message.value == b'\x00\x00\x00\x00\x15\x02\x00\xcc\xb8\xeb\xa6\x80_\x00\x00\x08test'



"""This module contains implementation of extended confluent-kafka Producer's API."""

import atexit
import traceback
from typing import Any, Union, Optional

from confluent_kafka import KafkaException

from wunderkafka.config.krb.rdkafka import challenge_krb_arg
from wunderkafka.logger import logger
from wunderkafka.types import DeliveryCallback
from wunderkafka.config import ProducerConfig
from wunderkafka.callbacks import error_callback
from wunderkafka.producers.abc import AbstractProducer
from wunderkafka.hotfixes.watchdog.types import Watchdog


class BytesProducer(AbstractProducer):
    """Producer implementation of extended interface for raw messages."""

    # FixMe (tribunsky.kir): add watchdog page reference
    def __init__(self, config: ProducerConfig, sasl_watchdog: Optional[Watchdog] = None) -> None:
        """
        Init producer.

        :param config:          Pydantic model with librdkafka producer's configuration.
        :param sasl_watchdog:   Callable to handle global state of kerberos auth (see Watchdog).
        """
        try:
            super().__init__(config.dict())
        except KafkaException as exc:
            config = challenge_krb_arg(exc, config)
            super().__init__(config.dict())

        self._config = config
        self._sasl_watchdog = sasl_watchdog
        atexit.register(self.flush)

    # ToDo (tribunsky.kir): make inherited from RDConfig models immutable.
    #                       Currently it explodes because of mutation in watchdog.
    #                       Do we need re-initiation of consumer/producer in runtime?
    @property
    def config(self) -> ProducerConfig:
        """
        Get the producer's config.

        :return:        Pydantic model with librdkafka producer's configuration.
        """
        return self._config

    def send_message(  # noqa: D102,WPS211  # inherited from superclass.
        self,
        topic: str,
        value: Optional[Union[str, bytes]] = None,  # noqa: WPS110  # Domain. inherited from superclass.
        key: Optional[Union[str, bytes]] = None,
        partition: Optional[int] = None,
        on_delivery: Optional[DeliveryCallback] = error_callback,
        *args: Any,
        blocking: bool = False,
        **kwargs: Any,
    ) -> None:
        if self._sasl_watchdog is not None:
            self._sasl_watchdog()
        if partition is not None:
            self.produce(topic, value, key=key, partition=partition, on_delivery=on_delivery, **kwargs)
        else:
            self.produce(topic, value, key=key, on_delivery=on_delivery, **kwargs)
        if blocking:
            self.flush()
        else:
            self.poll(0)



from typing import Any, Dict, List, Union, Optional

from wunderkafka.errors import SchemaRegistryLookupException
from wunderkafka.logger import logger
from wunderkafka.structures import SRMeta, SchemaMeta, ParsedHeader
from wunderkafka.schema_registry.abc import AbstractHTTPClient, AbstractSchemaRegistry
from wunderkafka.serdes.avro.headers import PROTOCOLS
from wunderkafka.schema_registry.cache import SimpleCache, AlwaysEmptyCache


def prepare_schemas_for_caching(
    versions: Dict[str, List[Dict[str, Optional[Union[int, str]]]]],
) -> Dict[ParsedHeader, str]:
    schemas_to_header_mapping = {}
    for schema in versions['entities']:
        txt = schema['schemaText']
        if not isinstance(txt, str):
            raise ValueError('Provided SchemaText is not str: {0}'.format(schema))
        meta_id = schema['schemaMetadataId']
        if not isinstance(meta_id, int):
            raise ValueError('Wrong schemaMetadataId (not int): {0}'.format(schema))
        schema_version = schema['version']
        if not isinstance(schema_version, int):
            raise ValueError('Wrong schema version (not int): {0}'.format(schema))
        new_header_v1 = ParsedHeader(
            protocol_id=1,
            meta_id=meta_id,
            schema_version=schema_version,
            schema_id=None,
            size=PROTOCOLS[1].header_size + 1,
        )
        schema_id = schema['id']
        if not isinstance(schema_id, int):
            raise ValueError('Wrong schema id (not int): {0}'.format(schema))
        new_header_v2 = ParsedHeader(
            protocol_id=2,
            meta_id=None,
            schema_version=None,
            schema_id=schema_id,
            size=PROTOCOLS[2].header_size + 1,
        )
        new_header_v3 = ParsedHeader(
            protocol_id=3,
            meta_id=None,
            schema_version=None,
            schema_id=schema_id,
            size=PROTOCOLS[3].header_size + 1,
        )
        for new_header in (new_header_v1, new_header_v2, new_header_v3):
            schemas_to_header_mapping[new_header] = txt
    return schemas_to_header_mapping


class ClouderaSRClient(AbstractSchemaRegistry):

    def __init__(self, http_client: AbstractHTTPClient, cache: Optional[SimpleCache] = None) -> None:
        self._client = http_client
        self._cache = AlwaysEmptyCache() if cache is None else cache
        self._requests_count = 0

    @property
    def requests_count(self) -> int:
        return self._requests_count

    def _send_request_for_schema(self, meta: SchemaMeta) -> Any:
        versions = self._client.make_request('schemas/{0}/versions'.format(meta.subject))
        self._requests_count += 1
        return versions

    def _create_meta(self, subject: str) -> int:
        body = {
            'type': 'avro',
            'schemaGroup': 'Kafka',
            'name': subject,
            'description': subject,
            'compatibility': 'BACKWARD',
            'validationLevel': 'ALL',
        }
        return self._client.make_request('schemas', method='POST', body=body)

    def _create_schema(self, subject: str, schema_text: str) -> int:
        body = {
            "description": subject,
            # TypeError: Object of type 'mappingproxy' is not JSON serializable
            # https://github.com/confluentinc/confluent-kafka-python/issues/610
            "schemaText": schema_text,
        }
        query = {'branch': 'MASTER'}
        return self._client.make_request('schemas/{0}/versions'.format(subject), method='POST', body=body, query=query)

    def get_schema_text(self, meta: SchemaMeta) -> str:
        # ToDo (tribunsky.kir): arguably, the best key is BINARY header, not dataclass.
        # ToDo (tribunsky.kir): ineffective caching. When we get response with list of schemas,
        #                       we can do better: pre-generate all possible metas and keep texts
        schema_text = self._cache.get(meta.header)
        if schema_text is not None:
            return schema_text
        logger.debug("Couldn't find header ({0}) in cache, re-requesting...".format(meta.header))
        versions = self._send_request_for_schema(meta)
        for hdr, txt in prepare_schemas_for_caching(versions).items():
            self._cache.set(hdr, txt)
        schema_text = self._cache.get(meta.header)
        if schema_text is None:
            error_message = "Couldn't find schema for {0}".format(meta.header)
            logger.error(error_message)
            for schema in versions['entities']:
                logger.warning('\tschema: {0}'.format(schema))
            raise SchemaRegistryLookupException(error_message)
        return schema_text

    # ToDo (tribunsky.kir): here and in producers/constructor:
    #                       - make hypothetical meta + prop for subject
    #                       - symmetry with get_schema_text
    def register_schema(self, topic: str, schema_text: str, *, is_key: bool = True) -> SRMeta:
        uid = (topic, schema_text, is_key)
        meta = self._cache.get(uid)
        if meta is None:
            suffix = ':k' if is_key else ''
            subject = '{0}{1}'.format(topic, suffix)
            # ToDo (tribunsky.kir): three queries is bad. Either we should know protocol id,
            #                       e.g. to skip last query (we do not need id always)
            meta_id = self._create_meta(subject)
            schema_version = self._create_schema(subject, schema_text)
            versions = self._client.make_request('schemas/{0}/versions'.format(subject))
            for mt in versions['entities']:
                if mt['schemaMetadataId'] == meta_id and mt['version'] == schema_version:
                    meta = SRMeta(
                        meta_id=meta_id,
                        schema_version=schema_version,
                        schema_id=mt['id']
                    )
            self._cache.set(uid, meta)
        return meta



"""This module contains implementation of extended confluent-kafka Consumer's API."""

import time
import atexit
import datetime
import traceback
from typing import Dict, List, Union, Optional

from confluent_kafka import KafkaException

from wunderkafka.config.krb.rdkafka import challenge_krb_arg
from wunderkafka.types import HowToSubscribe
from wunderkafka.config import ConsumerConfig
from wunderkafka.errors import ConsumerException
from wunderkafka.logger import logger
from wunderkafka.callbacks import reset_partitions
from wunderkafka.consumers.abc import Message, AbstractConsumer
from wunderkafka.consumers.subscription import TopicSubscription
from wunderkafka.hotfixes.watchdog.types import Watchdog


class BytesConsumer(AbstractConsumer):
    """Consumer implementation of extended interface for raw messages."""

    # FixMe (tribunsky.kir): add watchdog page reference
    def __init__(self, config: ConsumerConfig, sasl_watchdog: Optional[Watchdog] = None) -> None:
        """
        Init consumer.

        :param config:          Pydantic model with librdkafka consumer's configuration.
        :param sasl_watchdog:   Callable to handle global state of kerberos auth (see Watchdog).
        """
        try:
            super().__init__(config.dict())
        except KafkaException as exc:
            config = challenge_krb_arg(exc, config)
            super().__init__(config.dict())
        self.subscription_offsets: Optional[Dict[str, HowToSubscribe]] = None

        self._config = config
        self._last_poll_ts = time.perf_counter()
        self._sasl_watchdog = sasl_watchdog
        # ToDo (tribunsky-kir): make it configurable
        atexit.register(self.close)

    def __str__(self) -> str:
        """
        Get human-readable representation of consumer.

        :return:    string with consumer gid.
        """
        return '{0}:{1}'.format(self.__class__.__name__, self._config.group_id)

    def batch_poll(  # noqa: D102  # inherited from superclass.
        self,
        timeout: float = 1.0,
        num_messages: int = 1000000,
        *,
        raise_on_lost: bool = False,
    ) -> List[Message]:
        if self._sasl_watchdog is not None:
            self._sasl_watchdog()

        # ToDo (tribunsky.kir): naybe it better to use on_lost callback within subscribe()
        dt = int((time.perf_counter() - self._last_poll_ts) * 1000)
        if dt > self._config.max_poll_interval_ms:
            msg = 'Exceeded max.poll.interval.ms ({0}): {1}'.format(self._config.max_poll_interval_ms, dt)

            if raise_on_lost:
                # ToDo (tribunsky.kir): resubscribe by ts?
                raise ConsumerException(msg)
            logger.warning(msg)

        msgs = self.consume(num_messages=num_messages, timeout=timeout)
        self._last_poll_ts = time.perf_counter()
        return msgs

    # ToDo (tribunsky.kir): do not override original API and wrap it in superclass
    def subscribe(  # noqa: D102,WPS211  # inherited from superclass.
        self,
        topics: List[Union[str, TopicSubscription]],
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        subscriptions = {}
        for tpc in topics:
            if isinstance(tpc, str):
                tpc = TopicSubscription(
                    topic=tpc, from_beginning=from_beginning, offset=offset, ts=ts, with_timedelta=with_timedelta,
                )
            subscriptions[tpc.topic] = tpc.how

        # We have specific subscription at least once
        if any(subscriptions.values()):
            self.subscription_offsets = subscriptions
            # ToDo (tribunsky.kir): avoid mutation of self.subscription_offset and remove it as a field
            super().subscribe(topics=list(self.subscription_offsets), on_assign=reset_partitions)
        else:
            super().subscribe(topics=list(subscriptions))



import pytest
from pydantic import ValidationError

from tests.smoke.conftest import RawConfig
from wunderkafka import SRConfig, BytesConsumer, BytesProducer, ConsumerConfig, ProducerConfig


def test_init_consumer(boostrap_servers: str) -> None:
    config = ConsumerConfig(group_id='my_group', bootstrap_servers=boostrap_servers)
    consumer = BytesConsumer(config)
    print(consumer)

    with pytest.raises(AttributeError):
        consumer.config = ConsumerConfig(                                                                 # type: ignore
            group_id='my_other_group',
            bootstrap_servers=boostrap_servers,
        )
    assert consumer.config == config

    consumer.close()


def test_init_producer(boostrap_servers: str) -> None:
    config = ProducerConfig(bootstrap_servers=boostrap_servers)
    BytesProducer(config)


def test_init_no_krb_consumer(non_krb_config: RawConfig) -> None:
    config = ConsumerConfig(group_id='my_group', **non_krb_config)
    consumer = BytesConsumer(config)
    print(consumer)

    with pytest.raises(AttributeError):
        consumer.config = ConsumerConfig(                                                                 # type: ignore
            group_id='my_other_group',
            **non_krb_config,
        )
    assert consumer.config == config

    consumer.close()


def test_init_no_krb_producer(non_krb_config: RawConfig) -> None:
    config = ProducerConfig(**non_krb_config)
    BytesProducer(config)


def test_sr_required_url() -> None:
    with pytest.raises(ValidationError):
        SRConfig()


def test_group_id_required() -> None:
    with pytest.raises(ValidationError):
        ConsumerConfig()



from pathlib import Path

import pytest

from wunderkafka.tests import TestHTTPClient
from wunderkafka.errors import SchemaRegistryLookupException
from wunderkafka.structures import SchemaMeta, ParsedHeader
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient


@pytest.fixture
def client(sr_root: Path) -> ClouderaSRClient:
    return ClouderaSRClient(TestHTTPClient(sr_root), SimpleCache())


@pytest.fixture
def topic() -> str:
    return 'test_signals'


@pytest.fixture
def schema_v1(topic: str) -> SchemaMeta:
    return SchemaMeta(
        topic=topic,
        is_key=False,
        header=ParsedHeader(protocol_id=1, meta_id=1692, schema_id=None, schema_version=1, size=13),
    )


@pytest.fixture
def schema_v2(topic: str) -> SchemaMeta:
    return SchemaMeta(
        topic=topic,
        is_key=False,
        header=ParsedHeader(protocol_id=1, meta_id=1692, schema_id=None, schema_version=2, size=13),
    )


def test_runtime_evolving(client: ClouderaSRClient, schema_v1: SchemaMeta, schema_v2: SchemaMeta) -> None:
    assert client.get_schema_text(schema_v1)
    with pytest.raises(SchemaRegistryLookupException):
        client.get_schema_text(schema_v2)
    assert client.requests_count == 2



"""
Wunderkafka - the power of librdkafka for humans^W pythons.

Module contains handful re-imports (public API) from the package.

:copyright: (c) 2021 by Severstal Digital.
:license: Apache 2.0, see LICENSE for more details.
"""
# ToDo (tribunsky-kir): API doesn't belong to us, reconsider of exposing confluent_kafka's Message
from confluent_kafka import Message

from wunderkafka.config import ConsumerConfig, ProducerConfig
from wunderkafka.factories import (
    AvroConsumer,
    AvroProducer,
    AvroModelProducer,
    AvroClouderaConsumer,
    AvroConfluentConsumer,
    AvroModelClouderaProducer,
    AvroModelConfluentProducer,
)
from wunderkafka.protocols import AnyConsumer, AnyProducer
from wunderkafka.consumers.bytes import BytesConsumer
from wunderkafka.producers.bytes import BytesProducer
from wunderkafka.config.generated.enums import (
    IsolationLevel,
    AutoOffsetReset,
    CompressionType,
    QueuingStrategy,
    CompressionCodec,
    SecurityProtocol,
    BrokerAddressFamily,
    SslEndpointIdentificationAlgorithm,
)
from wunderkafka.config.schema_registry import SRConfig
from wunderkafka.consumers.subscription import TopicSubscription



from tests.smoke.conftest import RawConfig
from wunderkafka import ConsumerConfig, ProducerConfig
from wunderkafka.config.krb.rdkafka import config_requires_kerberos


def test_default_consumer_not_requires_krb(default_config: RawConfig) -> None:
    config = ConsumerConfig(group_id='my_group', **default_config)
    assert config_requires_kerberos(config) is False


def test_default_producer_not_requires_krb(default_config: RawConfig) -> None:
    config = ProducerConfig(**default_config)
    assert config_requires_kerberos(config) is False


def test_consumer_requires_krb(krb_config: RawConfig) -> None:
    config = ConsumerConfig(group_id='my_group', **krb_config)
    assert config_requires_kerberos(config) is True


def test_producer_requires_krb(krb_config: RawConfig) -> None:
    config = ProducerConfig(**krb_config)
    assert config_requires_kerberos(config) is True


def test_consumer_not_requires_krb(non_krb_config: RawConfig) -> None:
    config = ConsumerConfig(group_id='my_group', **non_krb_config)
    assert config_requires_kerberos(config) is False


def test_producer_not_requires_krb(non_krb_config: RawConfig) -> None:
    config = ProducerConfig(**non_krb_config)
    assert config_requires_kerberos(config) is False



from pathlib import Path

import pytest


@pytest.fixture
def topic() -> str:
    return 'testing_wunderkafka'


@pytest.fixture
def sr_root_create(fixtures_root: Path) -> Path:
    return fixtures_root / 'schema_registry' / 'confluent' / 'create'


@pytest.fixture
def sr_root_existing(fixtures_root: Path) -> Path:
    return fixtures_root / 'schema_registry' / 'confluent' / 'existing'


@pytest.fixture
def sr_root_update(fixtures_root: Path) -> Path:
    return fixtures_root / 'schema_registry' / 'confluent' / 'update'



"""This module contains aliases and helper definitions for type hints."""

# ToDo (tribunsky.kir): move it to module with structures
from typing import Any, Tuple, Union, Callable, Optional, Type

from confluent_kafka import Message, KafkaError

from wunderkafka.structures import Offset, SRMeta, Timestamp, ParsedHeader, SchemaDescription

# ToDo (tribunsky.kir): add symmetry and generalize this part
HeaderParser = Callable[[bytes], ParsedHeader]
HeaderPacker = Callable[[int, SRMeta], bytes]

DeliveryCallback = Callable[[Optional[KafkaError], Message], None]

MsgValue = Any
MsgKey = Any

KeySchemaDescription = SchemaDescription
ValueSchemaDescription = SchemaDescription

MessageDescription = Union[str, Tuple[str, str], Type, Tuple[Type, Type]]
TopicName = str

HowToSubscribe = Optional[Union[Timestamp, Offset]]



# -*- coding: utf-8 -*-

"""Some timing boilerplate."""

import time
import datetime


def now() -> int:
    """
    Return UNIX timestamp in ms.

    :returns:        current timestamp (ms)
    """
    return int(time.time()*1000)


def ts2dt(ts: float) -> datetime.datetime:
    """
    Convert unix-timestamp to datetime object.

    :param ts:              Timestamp in seconds or milliseconds.
    :returns:               Corresponding datetime object.

    :raises ValueError:     Raised when couldn't coerce float to datetime.
    """
    ts_str_len_ms = 13      # 1559574586123
    current_ts_len = len(str(int(ts)))
    if current_ts_len <= 10:  # 155957458
        return datetime.datetime.fromtimestamp(ts)
    if current_ts_len == ts_str_len_ms:
        return datetime.datetime.fromtimestamp(round(ts/1000))
    raise ValueError('Invalid timestamp: {0} seems not to be a UNIX-timestamp in seconds or milliseconds'.format(ts))



import pytest

from wunderkafka.hotfixes.watchdog import parse_kinit, KinitParams

USER = 'my.user'
REALM = 'MYDOMAIN.COM'


@pytest.mark.parametrize(
    'keytab',
    ['{0}.keytab'.format(USER), '/home/{0}@{1}/{0}.keytab'.format(USER, REALM.lower())]
)
def test_parse_kinit(keytab: str) -> None:
    kinit_cmd = 'kinit {0}@{1} -k -t {2}'.format(USER, REALM, keytab)
    assert parse_kinit(kinit_cmd) == KinitParams(user=USER, realm=REALM, keytab=keytab, cmd=kinit_cmd)



import os
import logging
import operator
from typing import Dict, List, Tuple, Union, Optional, NamedTuple
from pathlib import Path
from collections import defaultdict

logging.basicConfig(
    format='%(asctime)s - %(levelname)s - %(message)s',
    level=logging.DEBUG,
)
logger = logging

# mypy cannot into constrained type at the time of on-knee writing of this script
IGNORE_PYDANTIC_TYPE = '# type: ignore[valid-type]'

GROUP_ID = 'group.id'
SASL_MECHANISMS = 'sasl.mechanisms'

CLS_MAPPING = {
    'P': 'class RDProducerConfig(RDKafkaConfig):',
    'C': 'class RDConsumerConfig(RDKafkaConfig):',
    '*': 'class RDKafkaConfig(BaseSettings):',
}

TPL_MAPPING = {
    'P': 'PRODUCER_FIELDS = (',
    'C': 'CONSUMER_FIELDS = (',
    '*': 'COMMON_FIELDS = (',
}

TYPES_MAPPING = {
    'integer': 'int',
    'float': 'float',
    'string': 'str',
    'boolean': 'bool',
    'see dedicated API': 'Callable',
    'CSV flags': 'str',
    'pattern list': 'str',
}


class Row(NamedTuple):
    property_name: str
    property_belongs: str
    property_range: str
    property_default: str

    property_importance: str
    property_description: str

    @property
    def comment(self) -> bool:
        return '..' in self.property_range

    @property
    def deprecated(self) -> bool:
        # Markdown
        deprecated = '**deprecated'
        return self.property_description.casefold().startswith(deprecated)

    @property
    def name(self) -> str:
        return self.property_name.replace('.', '_')

    @property
    def cls(self) -> str:
        return CLS_MAPPING[self.property_belongs]

    @property
    def default(self) -> Optional[Union[str, bool, float]]:
        if not self.property_default:
            return None
        if self.type == 'boolean':
            return self.property_default.casefold() == 'true'
        if self.type == 'float':
            return float(self.property_default)
        if self.type in {'string', 'CSV flags'}:
            return "'{0}'".format(self.property_default)
        if self.type == 'enum value':
            class_name = ''.join([word.capitalize() for word in self.property_name.split('.')])
            return 'enums.{0}.{1}'.format(class_name, self.property_default)
        return self.property_default

    @property
    def range(self) -> Union[str, Tuple[str, str]]:
        delim = '..'
        if delim not in self.property_range:
            # search for bool
            pieces = [piece.strip().casefold() for piece in self.property_range.split(',') if piece.strip()]
            if set(pieces) == {'true', 'false'}:
                return 'bool'
            logger.warning("Couldn't treat range as bool either: {0}".format(self.property_range))
            return self.property_range
        pieces = [piece.strip() for piece in self.property_range.split(delim) if piece.strip()]
        if len(pieces) != 2:
            logger.warning("Couldn't convert range: {0}".format(self.property_range))
            return self.property_range
        ge, le = pieces
        return ge, le

    @property
    def annotation(self) -> str:
        if self.type == 'boolean':
            return TYPES_MAPPING[self.type]
        if not self.default:
            return 'Optional[{0}]'.format(TYPES_MAPPING[self.type])
        if self.type == 'string':
            return TYPES_MAPPING[self.type]
        if self.type == 'CSV flags':
            return TYPES_MAPPING[self.type]
        if '..' in self.property_range:
            ge, le = self.range                                                                           # type: ignore
            if self.type == 'integer':
                return 'conint(ge={0}, le={1})'.format(ge, le)
            elif self.type == 'float':
                return 'confloat(ge={0}, le={1})'.format(ge, le)
        if self.type == 'enum value':
            class_name = ''.join([word.capitalize() for word in self.property_name.split('.')])
            return 'enums.{0}'.format(class_name)
        err_msg = "You missed something, you, stupid! Check: {0} {1} {2}"
        raise ValueError(err_msg.format(self.name, self.property_range, self.default))

    @property
    def type(self) -> str:
        # logger.debug(self.property_description.split('Type: '))
        return self.property_description.split('Type: ')[-1].strip('*')

    def __str__(self) -> str:
        if self.property_name == GROUP_ID:
            return '    {0}: str'.format(self.name)
        if self.property_name == SASL_MECHANISMS:
            return '    # {0}: {1} = {2}'.format(self.name, self.annotation, self.default)
        if self.property_name == 'builtin.features':
            lines = ["', '.join(["]
            spaces = ' ' * 8
            for feat in self.default.strip('"').strip("'").split(','):                                    # type: ignore
                stripped = feat.strip()
                if stripped:
                    lines.append("{0}'{1}',".format(spaces, stripped))
            lines.append("    ])")
            return '    {0}: {1} = {2}'.format(self.name, self.annotation, '\n'.join(lines))
        if self.comment:
            left = '    {0}: {1} = {2}'.format(self.name, self.annotation, self.default)
            ws = 120 - len(left) - len(IGNORE_PYDANTIC_TYPE)
            if ws < 1:
                ws = 1
            return left + ws * ' ' + IGNORE_PYDANTIC_TYPE
        else:
            return '    {0}: {1} = {2}'.format(self.name, self.annotation, self.default)


def read_markdown(filename: Union[str, Path] = 'CONFIGURATION.md', *, cut: bool = False) -> List[str]:
    lines = []
    with open(filename, 'r') as fl:
        all_lines = [ln.strip() for ln in fl.read().split('\n') if ln.strip()]
        if not cut:
            return all_lines
    for line in all_lines:
        if line.startswith('## Topic configuration properties'):
            break
        lines.append(line)
    return lines


def write_python(lines: List[str], file_name: str) -> None:
    with open(file_name, 'w') as fl:
        for line in lines:
            fl.write('{0}\n'.format(line))


def parse_line(line: str) -> Optional[Row]:
    not_a_rows = ('#', '-', 'Property')
    for bad_start in not_a_rows:
        if line.startswith(bad_start):
            return None
    column_data = [piece.strip().replace('\|', '|') for piece in line.split(' | ')]
    if len(column_data) != 6:
        logger.warning("Couldn't parse: {0}".format(column_data))
        return None
    return Row(*column_data)


def parse(lines: List[str], *, allow_deprecated: bool = False) -> List[Row]:
    rows = []
    for line in lines:
        row = parse_line(line)
        if row is not None:
            if row.deprecated:
                if allow_deprecated is False:
                    logger.warning('Skipping {0} as deprecated'.format(row.property_name))
                else:
                    logger.warning('{0} is deprecated'.format(row.property_name))
                    rows.append(row)
            else:
                rows.append(row)
    logger.info('Total properties parsed: {0}'.format(len(rows)))
    return rows


def group(rows: List[Row]) -> Dict[str, List[Row]]:
    grps = defaultdict(list)
    for row in rows:
        grps[row.property_belongs].append(row)
    return grps


def generate_models(groups: Dict[str, List[Row]]) -> List[str]:
    properties = [
        '# I am not gonna to generate single type for every single range of conint/confloat.',
        '# https://github.com/samuelcolvin/pydantic/issues/156',
        '',
        'from typing import Callable, Optional',
        '',
        'from pydantic import conint, confloat',
        'from pydantic_settings import BaseSettings',
        '',
        "# Enums because we can't rely that client code uses linters.",
        '# Of course, it will fail with cimpl.KafkaException, but later, when Consumer/Producer are really initiated',
        'from wunderkafka.config.generated import enums'
    ]
    already_generated = set()
    for grp in sorted(groups):
        properties.append('')
        properties.append('')
        properties.append('{0}'.format(CLS_MAPPING[grp]))
        pre = []
        uniq = []
        for row in groups[grp]:
            if row.property_name in already_generated:
                logger.warning('Skipping generating second field ({0})'.format(row))
            else:
                if row.property_name == GROUP_ID:
                    pre.append(row)
                else:
                    uniq.append(row)
                already_generated.add(row.property_name)
        properties += [str(row) for row in sorted(pre, key=operator.attrgetter('name'))]
        for prop in sorted(uniq, key=operator.attrgetter('name')):
            if prop.property_name == SASL_MECHANISMS:
                properties.append('    # ToDo (tribunsky.kir): rethink using aliases? They may need simultaneous valdiation or may be injected via dict()')
                properties.append('    # It is just alias, but when setting it manually it may misbehave with current defaults.')
            properties.append(str(prop))

    return properties


def generate_fields(groups: Dict[str, List[Row]]) -> List[str]:
    properties = [
        "# Why so: not all configuration parameters of librdkafka may be easily replaced from '_' to '.',",
        "#   therefore, we can't convert on-the-fly from  `ssl_ca` without errors",
        "#   and we don't want to have a nice whitelist, which is arguable",
    ]
    for grp in sorted(groups):
        properties.append('{0}'.format(TPL_MAPPING[grp]))
        all_fields = sorted({row.property_name for row in groups[grp]})
        properties += ["    '{0}',".format(field) for field in all_fields]
        properties.append(')')
    return properties


def generate_enums(groups: Dict[str, List[Row]]) -> List[str]:
    properties = ['from enum import Enum']
    already_generated = set()
    for grp in sorted(groups):
        for row in groups[grp]:
            if row.type == 'enum value':
                if row.property_name in already_generated:
                    logger.warning('Skipping second enum for {0}'.format(row.property_name))
                else:
                    properties.append('')
                    properties.append('')
                    properties += generate_enum(row.property_name, row.property_range)
                    already_generated.add(row.property_name)
    return properties


def generate_enum(prop: str, rng: str) -> List[str]:
    cls_name = ''.join([word.capitalize() for word in prop.split('.')])
    header = 'class {0}(str, Enum):'.format(cls_name)
    flds = []
    for field in rng.split(','):
        flds.append("    {0} = '{0}'".format(field.strip()))
    return [header] + flds


Name = str
Version = str
Lines = List[str]
Files = Dict[Name, Lines]


def generate(lines: Dict[Version, Files]) -> Dict[Name, Lines]:
    dct: Dict[Name, Lines] = {}
    for file_name in ('enums.py', 'models.py', 'fields.py'):
        logger.info('Processing {0}'.format(file_name))
        s = set()
        for libversion in lines:
            content = '\n'.join(lines[libversion][file_name])
            s.add(content)
        if not len(s):
            raise ValueError('Nothing to compare')
        if len(s) == 1:
            logger.info('All generated files for {0} are equal, keeping just one'.format(file_name))
            dct[file_name] = lines[libversion][file_name]
        else:
            logger.warning("Generated files for {0} differs, merging it...".format(file_name))
            header = '# mypy: disable-error-code="no-redef"'
            if file_name == 'fields.py':
                header = '# mypy: disable-error-code="no-redef,assignment"'
            new_lines = [
                '# ToDo (tribunsky.kir): looks like that idea of dynamic import via imp depending on librdkafka',
                "#                       wasn't the worst idea, cause `if`s causes a lot of static checks problems.",
                header,
                '',
                'from wunderkafka import librdkafka',
                '',
            ]
            head, *tail = list(sorted(lines))
            new_lines.extend(lines[head][file_name])
            for libversion in tail:
                tpl = tuple(int(digit) for digit in libversion.split('.'))
                delimeter = [
                    '',
                    'if librdkafka.__version__ >= {0}:'.format(tpl)
                ]
                new_lines.extend(delimeter)
                new_lines.extend('    {0}'.format(line) for line in lines[libversion][file_name])
            new_lines.append('')
            dct[file_name] = new_lines
    return dct


def main() -> None:
    lines: Dict[Version, Files] = {}
    root_dir = Path(__file__).parent / 'versions'
    for sub_path in os.listdir(root_dir):
        path = root_dir / sub_path
        if path.is_dir():
            logger.info('{0}: handling...'.format(path))
            configuration_md = path / 'CONFIGURATION.md'
            grouped = group(parse(read_markdown(filename=configuration_md)))
            files = {
                'models.py': generate_models(grouped),
                'fields.py': generate_fields(grouped),
                'enums.py': generate_enums(grouped),
            }
            lines[sub_path] = files
        else:
            logger.info('{0}: skipping...'.format(path))
    dct = generate(lines)
    for file_name, content in dct.items():
        write_python(content, 'generated/{0}'.format(file_name))


def single() -> None:
    grouped = group(parse(read_markdown()))
    models = generate_models(grouped)
    write_python(models, 'models.py')
    fields = generate_fields(grouped)
    write_python(fields, 'fields.py')
    enums = generate_enums(grouped)
    write_python(enums, 'enums.py')


if __name__ == '__main__':
    main()



from typing import Type, Union, Optional
from pathlib import Path

from wunderkafka.types import TopicName, KeySchemaDescription, ValueSchemaDescription
from wunderkafka.serdes.abc import AbstractDescriptionStore
from dataclasses_avroschema import AvroModel
from wunderkafka.serdes.avromodel import derive


class SchemaTextRepo(AbstractDescriptionStore):

    def add(self, topic: TopicName, value: str, key: str) -> None:
        self._values[topic] = ValueSchemaDescription(text=value)
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=key)


def _load_from_file(filename: Path) -> str:
    with open(filename) as fl:
        return fl.read()


# ToDo (tribunsky.kir): refactor it, maybe add hooks to parent class.
#                       Barbara, forgive us. Looks like AbstractDescriptionStore should be generic.
class SchemaFSRepo(AbstractDescriptionStore):

    def add(self, topic: TopicName, value: Union[str, Path], key: Union[str, Path]) -> None:
        self._values[topic] = ValueSchemaDescription(text=_load_from_file(Path(value)))
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=_load_from_file(Path(key)))


class AvroModelRepo(AbstractDescriptionStore):

    def __init__(self) -> None:
        super().__init__()

    # ToDo (tribunsky.kir): change Type[AvroModel] to more general alias + check derivation from python built-ins
    def add(self, topic: TopicName, value: Type[AvroModel], key: Optional[Type[AvroModel]]) -> None:
        self._values[topic] = ValueSchemaDescription(text=derive(value, topic))
        if key is not None:
            self._keys[topic] = KeySchemaDescription(text=derive(key, topic, is_key=True))



from typing import Dict, List
from pathlib import Path

import pytest

from wunderkafka.config.generate import (
    Row,
    group,
    parse,
    read_markdown,
    generate_enums,
    generate_fields,
    generate_models,
)


def read_answer(root: Path, file_name: str) -> List[str]:
    with open(root / 'answers' / file_name) as fl:
        return fl.read().split('\n')[:-1]


@pytest.fixture
def fixture_root() -> Path:
    return Path(__file__).parent.parent / 'fixtures' / 'config'


@pytest.fixture
def configuration_md(fixture_root: Path) -> Path:
    return fixture_root / '1.5.0' / 'CONFIGURATION.md'


@pytest.fixture
def enums(fixture_root: Path) -> List[str]:
    return read_answer(fixture_root, 'enums.py')


@pytest.fixture
def fields(fixture_root: Path) -> List[str]:
    return read_answer(fixture_root, 'fields.py')


@pytest.fixture
def models(fixture_root: Path) -> List[str]:
    return read_answer(fixture_root, 'models.py')


@pytest.fixture
def grouped(configuration_md: Path) -> Dict[str, List[Row]]:
    return group(parse(read_markdown(filename=configuration_md)))


def test_enums(grouped: Dict[str, List[Row]], enums: List[str]) -> None:
    generated = generate_enums(grouped)
    assert enums == generated


def test_fields(grouped: Dict[str, List[Row]], fields: List[str]) -> None:
    generated = generate_fields(grouped)
    assert fields == generated


def test_models(grouped: Dict[str, List[Row]], models: List[str]) -> None:
    generated = generate_models(grouped)
    # Multiline for builtin
    assert models == '\n'.join(generated).split('\n')



from enum import Enum
from typing import Any

from requests.auth import AuthBase

try:
    import requests_kerberos
    from requests_kerberos import HTTPKerberosAuth
except ImportError:
    HAS_KERBEROS = False
else:
    HAS_KERBEROS = True

if HAS_KERBEROS:
    class HTTPKerberosMutualAuth(Enum):
        REQUIRED = requests_kerberos.REQUIRED
        OPTIONAL = requests_kerberos.OPTIONAL
        DISABLED = requests_kerberos.DISABLED

else:
    class HTTPKerberosMutualAuth(Enum):                                                                   # type: ignore
        REQUIRED = 1
        OPTIONAL = 2
        DISABLED = 3

    class HTTPKerberosAuth(AuthBase):                                                                     # type: ignore
        def __init__(self, *args: Any, **kwargs: Any) -> None:
            message = ' '.join([
                'Something went wrong: trying to use HTTPKerberosAuth while missing requests-kerberos.',
                'Maybe it is unexpected manual usage. Please, install requests-kerberos.'
            ])
            raise ImportError(message)



"""This module contains subscription helpers."""

import datetime
from typing import Union, Optional

from confluent_kafka import OFFSET_END, OFFSET_BEGINNING

from wunderkafka.structures import Offset, Timestamp


def choose_offset(
    from_beginning: Optional[bool] = None,
    offset: Optional[int] = None,
    ts: Optional[float] = None,
    with_timedelta: Optional[datetime.timedelta] = None,
) -> Optional[Union[Offset, Timestamp]]:
    """
    Choose subscription method (Offset, Timestamp, None), corresponding to given values.

    :param from_beginning:  If flag is set, return specific offset corresponding to beginning/end of the topic.
    :param offset:          If set, will return Offset object.
    :param ts:              If set, will return Timestamp.
    :param with_timedelta:  If set, will calculate timestamp for corresponding timedelta.

    :raises ValueError:     If multiple values are set: there is no sense in attempt to subscribe to topic via multiple
                            offsets (excluding multiple partitions scenario, which is not currently supported).

    :return:                None if no values are specified, Offset or Timestamp otherwise.
    """
    args = (from_beginning, offset, ts, with_timedelta)
    if sum([arg is not None for arg in args]) > 1:
        raise ValueError('Only one subscription method per time is allowed!')
    if from_beginning is not None:
        offset = OFFSET_BEGINNING if from_beginning else OFFSET_END
    if offset is not None:
        return Offset(offset)
    if with_timedelta is not None:
        start = datetime.datetime.now() - with_timedelta
        ts = start.timestamp() * 1000
    if ts is not None:
        return Timestamp(int(ts))
    return None


class TopicSubscription(object):
    """Allows custom definition of subscription for topic without need to build full TopicPartition list."""

    def __init__(  # noqa: WPS211  # ToDo (tribunsky.kir): reconsider API of 'how'
        self,
        topic: str,
        *,
        from_beginning: Optional[bool] = None,
        offset: Optional[int] = None,
        ts: Optional[int] = None,
        with_timedelta: Optional[datetime.timedelta] = None,
    ) -> None:
        """
        Init topic subscription object.

        Only one method of subscription per topic is allowed at time:

        - from beginning (depends on your retention policy)
        - from end (consume only latest messages)
        - via specific offset
        - via specific timestamp
        - via specific timedelta (from current datetime)
        - no special option (consumer will use "default" value of auto.offset.reset)

        :param topic:           Topic to subscribe.
        :param from_beginning:  If True, subscribe to get earliest available messages. If False, get latest messages.
        :param offset:          Subscribe to specific offset.
                                If offset not found, will behave with built-in default.
        :param ts:              Subscribe to specific timestamp (milliseconds).
                                If timestamp not found, will behave with built-in default.
        :param with_timedelta:  Subscribe to some moment in the past, from current datetime for a given timedelta.
                                Will calculate specific timestamp and subscribe via ts.
        """
        self.topic = topic
        self.how = choose_offset(
            from_beginning=from_beginning, offset=offset, ts=ts, with_timedelta=with_timedelta,
        )



from wunderkafka.serdes.avro.headers import ConfluentClouderaHeadersHandler
from wunderkafka.serdes.avro.serializers import FastAvroSerializer, AvroModelSerializer
from wunderkafka.serdes.avro.deserializers import FastAvroDeserializer



"""This module contains some ready-to-go combinations of the Consumer/Producer."""

from typing import Dict, Type, Union, Optional

from wunderkafka import ConsumerConfig, ProducerConfig
from wunderkafka.config.krb.rdkafka import config_requires_kerberos
from wunderkafka.types import TopicName, MessageDescription
from wunderkafka.serdes.avro import (
    FastAvroSerializer,
    AvroModelSerializer,
    FastAvroDeserializer,
    ConfluentClouderaHeadersHandler,
)
from wunderkafka.serdes.store import AvroModelRepo, SchemaTextRepo
from wunderkafka.consumers.bytes import BytesConsumer
from wunderkafka.producers.bytes import BytesProducer
from wunderkafka.schema_registry import SimpleCache, ClouderaSRClient, KerberizableHTTPClient
from wunderkafka.hotfixes.watchdog import check_watchdog
from wunderkafka.consumers.constructor import HighLevelDeserializingConsumer
from wunderkafka.producers.constructor import HighLevelSerializingProducer
from wunderkafka.schema_registry.clients.confluent import ConfluentSRClient


class AvroConsumer(HighLevelDeserializingConsumer):
    """Kafka Consumer client to get AVRO-serialized messages from Confluent/Cloudera installation."""

    def __init__(
        self,
        config: ConsumerConfig,
        *,
        sr_client: Optional[Union[Type[ClouderaSRClient], Type[ConfluentSRClient]]] = None,
    ) -> None:
        """
        Init consumer from pre-defined blocks.

        :param config:      configuration for:
                                - librdkafka consumer.
                                - schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry

        :raises ValueError: if schema registry configuration is missing.
        """
        sr = config.sr
        self._default_timeout: int = 60
        if sr is None:
            raise ValueError('Schema registry config is necessary for {0}'.format(self.__class__.__name__))
        if sr_client is None:
            sr_client = ClouderaSRClient

        config, watchdog = check_watchdog(config)
        super().__init__(
            consumer=BytesConsumer(config, watchdog),
            schema_registry=sr_client(
                KerberizableHTTPClient(sr, requires_kerberos=config_requires_kerberos(config)),
                SimpleCache(),
            ),
            headers_handler=ConfluentClouderaHeadersHandler().parse,
            deserializer=FastAvroDeserializer(),
        )


class AvroProducer(HighLevelSerializingProducer):
    """Kafka Producer client to serialize and send dictionaries or built-in types as messages."""

    def __init__(
        self,
        mapping: Optional[Dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        *,
        sr_client: Optional[Union[Type[ClouderaSRClient], Type[ConfluentSRClient]]] = None,
        protocol_id: int = 1
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value schema to be used fo serialization.
        :param config:      configuration for:
                                - librdkafka producer.
                                - schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry
        :param protocol_id: Protocol id for producer (1 - Cloudera, 0 - Confluent, etc.)

        :raises ValueError: if schema registry configuration is missing.
        """
        sr = config.sr
        if sr is None:
            raise ValueError('Schema registry config is necessary for {0}'.format(self.__class__.__name__))
        if sr_client is None:
            sr_client = ClouderaSRClient
        self._default_timeout: int = 60

        config, watchdog = check_watchdog(config)
        super().__init__(
            producer=BytesProducer(config, watchdog),
            schema_registry=sr_client(
                KerberizableHTTPClient(sr, requires_kerberos=config_requires_kerberos(config)),
                SimpleCache(),
            ),
            header_packer=ConfluentClouderaHeadersHandler().pack,
            serializer=FastAvroSerializer(),
            store=SchemaTextRepo(),
            mapping=mapping,
            protocol_id=protocol_id
        )


class AvroModelProducer(HighLevelSerializingProducer):
    """Kafka Producer client to serialize and send models or dataclasses as messages."""

    def __init__(
        self,
        mapping: Optional[Dict[TopicName, MessageDescription]],
        config: ProducerConfig,
        *,
        sr_client: Optional[Union[Type[ClouderaSRClient], Type[ConfluentSRClient]]] = None,
        protocol_id: int = 1
    ) -> None:
        """
        Init producer from pre-defined blocks.

        :param mapping:     Topic-to-Schemas mapping.
                            Mapping's value should contain at least message's value model to derive schema which will
                            be used fo serialization.
        :param config:      configuration for:
                                - librdkafka producer.
                                - schema registry client (conventional options for HTTP).

                            Refer original CONFIGURATION.md (https://git.io/JmgCl) or generated config.

        :param sr_client:   Client for schema registry
        :param protocol_id: Protocol id for producer (1 - Cloudera, 0 - Confluent, etc.)

        :raises ValueError: if schema registry configuration is missing.
        """
        sr = config.sr
        if sr is None:
            raise ValueError('Schema registry config is necessary for {0}'.format(self.__class__.__name__))

        if sr_client is None:
            sr_client = ClouderaSRClient
        self._default_timeout: int = 60

        config, watchdog = check_watchdog(config)
        super().__init__(
            producer=BytesProducer(config, watchdog),
            schema_registry=sr_client(
                KerberizableHTTPClient(sr, requires_kerberos=config_requires_kerberos(config)),
                SimpleCache(),
            ),
            header_packer=ConfluentClouderaHeadersHandler().pack,
            serializer=AvroModelSerializer(),
            store=AvroModelRepo(),
            mapping=mapping,
            protocol_id=protocol_id
        )


class AvroConfluentConsumer(AvroConsumer):
    """Kafka Consumer client to get AVRO-serialized messages from Confluent installation."""

    def __init__(self, config: ConsumerConfig):
        super().__init__(config, sr_client=ConfluentSRClient)


class AvroClouderaConsumer(AvroConsumer):
    """Kafka Consumer client to get AVRO-serialized messages from Cloudera installation."""

    def __init__(self, config: ConsumerConfig):
        super().__init__(config, sr_client=ClouderaSRClient)


class AvroModelConfluentProducer(AvroModelProducer):
    """Kafka Confluent Producer client to serialize and send models or dataclasses as messages."""

    def __init__(self, mapping: Optional[Dict[TopicName, MessageDescription]], config: ProducerConfig):
        super().__init__(mapping, config, sr_client=ConfluentSRClient, protocol_id=0)


class AvroModelClouderaProducer(AvroModelProducer):
    """Kafka Cloudera Producer client to serialize and send models or dataclasses as messages."""

    def __init__(self, mapping: Optional[Dict[TopicName, MessageDescription]], config: ProducerConfig):
        super().__init__(mapping, config, sr_client=ClouderaSRClient, protocol_id=1)
